#include <math.h>

inline void dd_add_dd(double *chi, double *clo, double ahi, double alo, double bhi, double blo);
inline void s_add_dd(double *chi, double *clo, double a, double bhi, double blo);
inline void dd_sub_dd(double *chi, double *clo, double ahi, double alo, double bhi, double blo);
inline void s_sub_dd(double *chi, double *clo, double a, double bhi, double blo);
inline void dd_mul_dd(double *chi, double *clo, double ahi, double alo, double bhi, double blo);
inline void dd_mul_s(double *chi, double *clo, double ahi, double alo, double b);
inline void s_mul_dd(double *chi, double *clo, double a, double bhi, double blo);
inline void dd_sqr(double *chi, double *clo, double ahi, double alo);
inline void dd_div_dd(double *chi, double *clo, double ahi, double alo, double bhi, double blo);
inline void ddsqrt_scl(double *chi, double *clo, double ahi, double alo);
inline void sin_taylor(double *chi, double *clo, double ahi, double alo);
inline void cos_taylor(double *chi, double *clo, double ahi, double alo);
inline void sincos_taylor(double *shi, double *slo, double *chi, double *clo, double ahi, double alo);
inline void nint(double *chi, double *clo, double ahi, double alo);


//******************************************************************************************//
// including files
//   ddsqrt     : square root
//   ddsin      : sine
//   ddcos      : cosine
//   ddtan      : tangent
//   ddexp      : exponent
//   basic arithmetic for dd
//
// last modified : 2012/03/09
//
//******************************************************************************************//


//--------------------------------------------------------------------------------------------
//
// double-double square root
//
// args
// ahi, alo : double numbers
// ahi + alo = dd number
// *m : row
// *n : column
// return (chi, clo) for dd number
//
//--------------------------------------------------------------------------------------------

int ddsqrt(double *chi, double *clo, double *ahi, double *alo, int *m, int *n)
{
	double c,cc,ah,al,p1,p2;
	int i;
	
	for(i=0; i<(*m)*(*n); i++) {
		chi[i] = 0.0;
		clo[i] = 0.0;
		
		c = sqrt(ahi[i]);
		p1 = c*c;
		cc = 134217729 * c;
		ah = cc - (cc - c);
		al = c - ah;
		p2 = ((ah * ah - p1) + 2.0 * ah * al) + al * al;
		
		cc = (((ahi[i] - p1) - p2) + alo[i]) / (2.0 * c);
		
		chi[i] = c + cc;
		clo[i] = cc - (chi[i] - c);
	}
	return(0);
}


//--------------------------------------------------------------------------------------------
//
// double-double sine
//
// args
// xhi, xlo : double numbers
// xhi + xlo = dd number
//
// return (shi, slo) for dd number
//
//--------------------------------------------------------------------------------------------
int ddsin(double *shi, double *slo, double *xhi, double *xlo) {
	
	double zh, zl,zhi, zlo, rh, rl, rhi, rlo, j, th, tl, thi, tlo, k,_2pih, _2pil, _pi2h, _pi2l, _pi16h, _pi16l;
	double uhi, ulo, vhi, vlo, sin_th, sin_tl, cos_th, cos_tl;	
	double c1=0.0,	c2=0.0,	c3=0.0,	c4=0.0;	
	int abs_k=0, int_j;
	
	double sin_table[2][4] = {
	{1.950903220161282758e-1, 3.826834323650897818e-1, 5.555702330196021776e-1, 7.071067811865475727e-1},
	{-7.991079068461731263e-18, -1.005077269646158761e-17, 4.709410940561676821e-17, -4.833646656725475727e-17}};

	double cos_table[2][4] = {
	{9.807852804032304306e-1, 9.238795325112867385e-1, 8.314696123025452357e-1, 7.071067811865475727e-1},
	{1.854693999782500573e-17, 1.764504708433667706e-17, 1.407385698472802389e-18, -4.833646656726456726e-17}};
	
	if((xhi[0]==0)&&(xlo[0]==0)) {
		shi[0]=0.0;
		slo[0]=0.0;
		return(0);
	}
	
	_2pih = 6.283185307179586232e+0;
	_2pil = 2.449293598294706414e-16;
	_pi2h = 1.570796326794896558e+0;
	_pi2l = 6.123233995736766036e-17;
	_pi16h= 1.963495408493620697e-1;
	_pi16l= 7.654042494670957545e-18;

	dd_div_dd(&zh, &zl, xhi[0], xlo[0], _2pih, _2pil);		//z'=x/2pi
	nint(&zhi, &zlo, zh, zl);		//z=nint(z')
	dd_mul_dd(&rh, &rl, _2pih, _2pil, zhi, zlo);		//r'=2pi*z
	dd_sub_dd(&rhi, &rlo, xhi[0], xlo[0], rh, rl);	//r=x-r';
	
	
	j=floor(rhi/_pi2h + 0.5);
	dd_mul_s(&th, &tl, _pi2h, _pi2l, j);	//t'=(pi/2)*j
	dd_sub_dd(&thi,&tlo, rhi, rlo, th, tl);	//t=r-t';
	
	k=floor(thi/_pi16h+0.5);
	dd_mul_s(&th, &tl, _pi16h, _pi16l, k);	//t'=(pi/16)*k
	dd_sub_dd(&thi,&tlo, thi, tlo, th, tl); //t=t*t';
	
	
	abs_k=(int)fabs(k);	//abs_k=abs(k)
	
	if((j<-2) || (j>2)) {
		shi[0]=0.0;
		slo[0]=10.0;
	}
	
	if(abs_k>4) {
		shi[0]=0.0;
		slo[0]=20.0;
	}
	
	int_j=(int)j;
	if(k==0) {
		switch(int_j) {
			case 0:
				sin_taylor(&shi[0], &slo[0], thi, tlo);
				return(0);	
			case 1:
				cos_taylor(&shi[0], &slo[0], thi, tlo);
				return(0);
			case -1:
				cos_taylor(&shi[0], &slo[0], thi, tlo);
				shi[0]=-shi[0];
				slo[0]=-slo[0];
				return(0);
			case 2:
			case -2:
				sin_taylor(&shi[0], &slo[0], thi, tlo);
				shi[0]=-shi[0];
				slo[0]=-slo[0];
				return(0);
		}
	}
	
	uhi=cos_table[0][abs_k-1];
	ulo=cos_table[1][abs_k-1];
	vhi=sin_table[0][abs_k-1];
	vlo=sin_table[1][abs_k-1];
	sincos_taylor(&sin_th, &sin_tl, &cos_th, &cos_tl, thi, tlo);
	
	if(j==0) {
		if(k>0) {
			dd_mul_dd(&c1, &c2, uhi, ulo, sin_th, sin_tl);
			dd_mul_dd(&c3, &c4, vhi, vlo, cos_th, cos_tl);
			dd_add_dd(&shi[0], &slo[0], c1, c2, c3, c4);
		}
		else {
			dd_mul_dd(&c1, &c2, uhi, ulo, sin_th, sin_tl);
			dd_mul_dd(&c3, &c4, vhi, vlo, cos_th, cos_tl);
			dd_sub_dd(&shi[0], &slo[0], c1, c2, c3, c4);
		}
	}
	else if(j==1) {
		if(k>0) {
			dd_mul_dd(&c1, &c2, uhi, ulo, cos_th, cos_tl);
			dd_mul_dd(&c3, &c4, vhi, vlo, sin_th, sin_tl);
			dd_sub_dd(&shi[0], &slo[0], c1, c2, c3, c4);
		}
		else {
			dd_mul_dd(&c1, &c2, uhi, ulo, cos_th, cos_tl);
			dd_mul_dd(&c3, &c4, vhi, vlo, sin_th, sin_tl);
			dd_add_dd(&shi[0], &slo[0], c1, c2, c3, c4);
		}
	}
	else if(j==-1) {
		if(k>0) {
			dd_mul_dd(&c1, &c2, vhi, vlo, sin_th, sin_tl);
			dd_mul_dd(&c3, &c4, uhi, ulo, cos_th, cos_tl);
			dd_sub_dd(&shi[0], &slo[0], c1, c2, c3, c4);	
		}
		else {
			dd_mul_dd(&c1, &c2, uhi, ulo, cos_th, cos_tl);
			dd_mul_dd(&c3, &c4, vhi, vlo, sin_th, sin_tl);
			c1=-c1;
			c2=-c2;
			dd_sub_dd(&shi[0], &slo[0], c1, c2, c3, c4);
		}
	}
	else {
		if(k>0) {
			dd_mul_dd(&c1, &c2, uhi, ulo, sin_th, sin_tl);
			c1=-c1;
			c2=-c2;
			dd_mul_dd(&c3, &c4, vhi, vlo, cos_th, cos_tl);
			dd_sub_dd(&shi[0], &slo[0], c1, c2, c3, c4);
		}
		else {
			dd_mul_dd(&c1, &c2, uhi, ulo, sin_th, sin_tl);
			dd_mul_dd(&c3, &c4, vhi, vlo, cos_th, cos_tl);
			dd_sub_dd(&shi[0], &slo[0], c3, c4, c1, c2);
		}
	}
	return(0);
}

//--------------------------------------------------------------------------------------------
//
// double-double cosine
//
// args
// xhi, xlo : double numbers
// xhi + xlo = dd number
//
// return (chi, clo) for dd number
//
//--------------------------------------------------------------------------------------------
int ddcos(double *chi, double *clo, double *xhi, double *xlo) {
	
	double zh, zl,zhi, zlo, rh, rl, rhi, rlo, j, th, tl, thi, tlo, k,_2pih, _2pil, _pi2h, _pi2l, _pi16h, _pi16l;
	double uhi, ulo, vhi, vlo, sin_th, sin_tl, cos_th, cos_tl;	
	double c1=0.0,c2=0.0,c3=0.0,c4=0.0;	
	int abs_k=0, int_j;
	
	double sin_table[2][4] = {
	{1.950903220161282758e-1, 3.826834323650897818e-1, 5.555702330196021776e-1, 7.071067811865475727e-1},
	{-7.991079068461731263e-18, -1.005077269646158761e-17, 4.709410940561676821e-17, -4.833646656725475727e-17}};

	double cos_table[2][4] = {
	{9.807852804032304306e-1, 9.238795325112867385e-1, 8.314696123025452357e-1, 7.071067811865475727e-1},
	{1.854693999782500573e-17, 1.764504708433667706e-17, 1.407385698472802389e-18, -4.833646656726456726e-17}};
	
	if((xhi[0]==0)&&(xlo[0]==0)) {
		chi[0]=1.0;
		clo[0]=0.0;
		return(0);
	}
	
	_2pih = 6.283185307179586232e+0;
	_2pil = 2.449293598294706414e-16;
	_pi2h = 1.570796326794896558e+0;
	_pi2l = 6.123233995736766036e-17;
	_pi16h= 1.963495408493620697e-1;
	_pi16l= 7.654042494670957545e-18;
    
	dd_div_dd(&zh, &zl, xhi[0], xlo[0], _2pih, _2pil);		//z'=x/2pi
	nint(&zhi, &zlo, zh, zl);		//z=nint(z')
	dd_mul_dd(&rh, &rl, _2pih, _2pil, zhi, zlo);		//r'=2pi*z
	dd_sub_dd(&rhi, &rlo, xhi[0], xlo[0], rh, rl);	//r=x-r';
	
	j=floor(rhi/_pi2h + 0.5);
	dd_mul_s(&th, &tl, _pi2h, _pi2l, j);	//t'=(pi/2)*j
	dd_sub_dd(&thi,&tlo, rhi, rlo, th, tl);	//t=r-t';
	
	k=floor(thi/_pi16h + 0.5);
	dd_mul_s(&th, &tl, _pi16h, _pi16l, k);	//t'=(pi/16)*k
	dd_sub_dd(&thi,&tlo, thi, tlo, th, tl); //t=t*t';
	
	
	abs_k=(int)fabs(k);	//abs_k=abs(k)
	
	if((j<-2) || (j>2)) {
		chi[0]=0.0;
		clo[0]=10.0;
	}
	
	if(abs_k>4) {
		chi[0]=0.0;
		clo[0]=20.0;
	}
	
	int_j=(int)j;
	if(k==0) {
		switch(int_j) {
			case 0:
				cos_taylor(&chi[0], &clo[0], thi, tlo);
				return(0);
			case 1:
				sin_taylor(&chi[0], &clo[0], thi, tlo);
				chi[0]=-chi[0];
				clo[0]=-clo[0];				
				return(0);
			case -1:
				sin_taylor(&chi[0], &clo[0], thi, tlo);
				return(0);
			case 2:
			case -2:
				cos_taylor(&chi[0], &clo[0], thi, tlo);
				chi[0]=-chi[0];
				clo[0]=-clo[0];
				return(0);
		}
	}
	
	uhi=cos_table[0][abs_k-1];
	ulo=cos_table[1][abs_k-1];
	vhi=sin_table[0][abs_k-1];
	vlo=sin_table[1][abs_k-1];
	sincos_taylor(&sin_th, &sin_tl, &cos_th, &cos_tl, thi, tlo);
	
	if(j==0) {
		if(k>0) {
			dd_mul_dd(&c1, &c2, uhi, ulo, cos_th, cos_tl);
			dd_mul_dd(&c3, &c4, vhi, vlo, sin_th, sin_tl);
			dd_sub_dd(&chi[0], &clo[0], c1, c2, c3, c4);
		}
		else {
			dd_mul_dd(&c1, &c2, uhi, ulo, cos_th, cos_tl);
			dd_mul_dd(&c3, &c4, vhi, vlo, sin_th, sin_tl);
			dd_add_dd(&chi[0], &clo[0], c3, c4, c1, c2);
		}
	}
	else if(j==1) {
		if(k>0) {
			dd_mul_dd(&c1, &c2, uhi, ulo, sin_th, sin_tl);
			dd_mul_dd(&c3, &c4, vhi, vlo, cos_th, cos_tl);
			c1=-c1;
			c2=-c2;
			dd_sub_dd(&chi[0], &clo[0], c1, c2, c3, c4);
		}
		else {
			dd_mul_dd(&c1, &c2, uhi, ulo, sin_th, sin_tl);
			dd_mul_dd(&c3, &c4, vhi, vlo, cos_th, cos_tl);
			c1=-c1;
			c2=-c2;			
			dd_add_dd(&chi[0], &clo[0], c1, c2, c3, c4);
		}
	}
	else if(j==-1) {
		if(k>0) {
			dd_mul_dd(&c1, &c2, uhi, ulo, sin_th, sin_tl);
			dd_mul_dd(&c3, &c4, vhi, vlo, cos_th, cos_tl);
			dd_add_dd(&chi[0], &clo[0], c1, c2, c3, c4);	
		}
		else {
			dd_mul_dd(&c1, &c2, uhi, ulo, sin_th, sin_tl);
			dd_mul_dd(&c3, &c4, vhi, vlo, cos_th, cos_tl);
			dd_sub_dd(&chi[0], &clo[0], c1, c2, c3, c4);
		}
	}
	else {
		if(k>0) {
			dd_mul_dd(&c1, &c2, uhi, ulo, cos_th, cos_tl);
			c1=-c1;
			c2=-c2;
			dd_mul_dd(&c3, &c4, vhi, vlo, sin_th, sin_tl);
			dd_add_dd(&chi[0], &clo[0], c1, c2, c3, c4);
		}
		else {
			dd_mul_dd(&c1, &c2, uhi, ulo, cos_th, cos_tl);
			c1=-c1;
			c2=-c2;			
			dd_mul_dd(&c3, &c4, vhi, vlo, sin_th, sin_tl);
			dd_sub_dd(&chi[0], &clo[0], c1, c2, c3, c4);
		}
	}
	return(0);
}


//--------------------------------------------------------------------------------------------
//
// double-double tangent
//
// args
// xhi, xlo : double numbers
// xhi + xlo = dd number
//
// return (tan_hi, tan_lo) for dd number
//
//--------------------------------------------------------------------------------------------
int ddtan(double *tan_hi, double *tan_lo, double *xhi, double *xlo) {
	
	double zh, zl,zhi, zlo, rh, rl, rhi, rlo, j, th, tl, thi, tlo, k,_2pih, _2pil, _pi2h, _pi2l, _pi16h, _pi16l;
	double uhi, ulo, vhi, vlo, sin_th, sin_tl, cos_th, cos_tl, shi, slo, chi, clo;	
	double c1=0.0,c2=0.0,c3=0.0,c4=0.0;	
	double sin_xh, sin_xl, cos_xh, cos_xl;	
	int abs_k=0, int_j, abs_j;
	
	double sin_table[2][4] = {
	{1.950903220161282758e-1, 3.826834323650897818e-1, 5.555702330196021776e-1, 7.071067811865475727e-1},
	{-7.991079068461731263e-18, -1.005077269646158761e-17, 4.709410940561676821e-17, -4.833646656725475727e-17}};

	double cos_table[2][4] = {
	{9.807852804032304306e-1, 9.238795325112867385e-1, 8.314696123025452357e-1, 7.071067811865475727e-1},
	{1.854693999782500573e-17, 1.764504708433667706e-17, 1.407385698472802389e-18, -4.833646656726456726e-17}};
	
	if((xhi[0]==0)&&(xlo[0]==0)) {
		tan_hi[0]=0.0;
		tan_lo[0]=0.0;
		return(0);
	}
	
	_2pih = 6.283185307179586232e+0;
	_2pil = 2.449293598294706414e-16;
	_pi2h = 1.570796326794896558e+0;
	_pi2l = 6.123233995736766036e-17;
	_pi16h= 1.963495408493620697e-1;
	_pi16l= 7.654042494670957545e-18;

	dd_div_dd(&zh, &zl, xhi[0], xlo[0], _2pih, _2pil);		//z'=x/2pi
	nint(&zhi, &zlo, zh, zl);		//z=nint(z')
	dd_mul_dd(&rh, &rl, _2pih, _2pil, zhi, zlo);		//r'=2pi*z
	dd_sub_dd(&rhi, &rlo, xhi[0], xlo[0], rh, rl);	//r=x-r';
	
	j=floor(rhi/_pi2h + 0.5);
	dd_mul_s(&th, &tl, _pi2h, _pi2l, j);	//t'=(pi/2)*j
	dd_sub_dd(&thi,&tlo, rhi, rlo, th, tl);	//t=r-t';
	
	k=floor(thi/_pi16h+0.5);
	dd_mul_s(&th, &tl, _pi16h, _pi16l, k);	//t'=(pi/16)*k
	dd_sub_dd(&thi,&tlo, thi, tlo, th, tl); //t=t*t';
	
	abs_j=(int)fabs(j);	//ans_j=abs(j)
	abs_k=(int)fabs(k);	//abs_k=abs(k)
	
	if((j<-2) || (j>2)) {
		tan_hi[0]=0.0;
		tan_lo[0]=10.0;
	}
	
	if(abs_k>4) {
		tan_hi[0]=0.0;
		tan_lo[0]=20.0;
	}

	sincos_taylor(&sin_th, &sin_tl, &cos_th, &cos_tl, thi, tlo);

	if(abs_k==0) {
		shi=sin_th;
		slo=sin_tl;
		chi=cos_th;
		clo=cos_tl;
	}
	else {
		uhi=cos_table[0][abs_k-1];
		ulo=cos_table[1][abs_k-1];
		vhi=sin_table[0][abs_k-1];
		vlo=sin_table[1][abs_k-1];
		if(k>0) {
			dd_mul_dd(&c1, &c2, uhi, ulo, sin_th, sin_tl);
			dd_mul_dd(&c3, &c4, vhi, vlo, cos_th, cos_tl);
			dd_add_dd(&shi, &slo, c1, c2, c3, c4);
			dd_mul_dd(&c1, &c2, uhi, ulo, cos_th, cos_tl);
			dd_mul_dd(&c3, &c4, vhi, vlo, sin_th, sin_tl);
			dd_sub_dd(&chi, &clo, c1, c2, c3, c4);
		}		
		else {
			dd_mul_dd(&c1, &c2, uhi, ulo, sin_th, sin_tl);
			dd_mul_dd(&c3, &c4, vhi, vlo, cos_th, cos_tl);
			dd_sub_dd(&shi, &slo, c1, c2, c3, c4);
			dd_mul_dd(&c1, &c2, uhi, ulo, cos_th, cos_tl);
			dd_mul_dd(&c3, &c4, vhi, vlo, sin_th, sin_tl);
			dd_add_dd(&chi, &clo, c1, c2, c3, c4);
		}
	}
	
	if(abs_j == 0) {
		sin_xh=shi;
		sin_xl=slo;
		cos_xh=chi;
		cos_xl=clo;
	}
	else if (j==1) {
		sin_xh=chi;
		sin_xl=clo;
		cos_xh=-shi;
		cos_xl=-slo;
	}
	else if (j==-1) {
		sin_xh=-chi;
		sin_xl=-clo;
		cos_xh=shi;
		cos_xl=slo;
	}
	else {
		sin_xh=-shi;
		sin_xl=-slo;
		cos_xh=-chi;
		cos_xl=-clo;
	}

	dd_div_dd(&tan_hi[0], &tan_lo[0], sin_xh, sin_xl, cos_xh, cos_xl);
	return(0);
}


//--------------------------------------------------------------------------------------------
//
// double-double exponent
//
// args
// xhi, xlo : double numbers
// xhi + xlo = dd number
//
// return (ehi, elo) for dd number
//
//--------------------------------------------------------------------------------------------
int ddexp(double *ehi, double *elo, double *xhi, double *xlo) {
	
	double k=512.0, inv_k=1.0/k;
	double log2_hi=6.931471805599452862e-01, log2_lo=2.319046813846299558e-17;
	double m, ph,pl, qh,ql, rh,rl;
	double sh,sl, vh,vl, th,tl;
	int i;
	double t=1.0;
	double eps=4.93038065763132e-32, thresh=inv_k*eps;
	double inv_fact[15][2] = {
        {1.66666666666666657e-01,  9.25185853854297066e-18},
        {4.16666666666666644e-02,  2.31296463463574266e-18},
        {8.33333333333333322e-03,  1.15648231731787138e-19},
        {1.38888888888888894e-03, -5.30054395437357706e-20},
        {1.98412698412698413e-04,  1.72095582934207053e-22},
        {2.48015873015873016e-05,  2.15119478667758816e-23},
        {2.75573192239858925e-06, -1.85839327404647208e-22},
        {2.75573192239858883e-07,  2.37677146222502973e-23},
        {2.50521083854417202e-08, -1.44881407093591197e-24},
        {2.08767569878681002e-09, -1.20834505911325997e-25},
        {1.60590438368216133e-10,  1.25852945887520981e-26},
        {1.14707455977297245e-11,  2.06555127528307454e-28},
        {7.64716373181981641e-13,  7.03872877733453001e-30},
        {4.77947733238738525e-14,  4.39920548583408126e-31},
        {2.81145725434552060e-15,  1.65088427308614326e-31}};
    
    

	if((xhi[0]==0.0) && (xlo[0]==0.0)) {
		ehi[0] = 1.0;
		elo[0] = 0.0;
		return(0);
	}

	if((xhi[0]==1.0) && (xlo[0]==0.0)) {
		ehi[0] = 2.7182818284590451;
		elo[0] = 1.4456468917292502e-16;
		return(0);
	}

	if(xhi[0]<=-709) {		//underflow
		ehi[0] = 0.0;
		elo[0] = 0.0;
		return(0);
	}

	if(xhi[0]>=709) {		//overflow
		ehi[0] = 0.0;
		elo[0] = 1.0;
		return(0);
	}

	m = floor(xhi[0] / log2_hi + 0.5);

	dd_mul_s(&ph, &pl, log2_hi, log2_lo, m);
	dd_sub_dd(&qh, &ql, xhi[0], xlo[0], ph, pl);
	rh = qh * inv_k;
    rl = ql * inv_k;

    dd_mul_dd(&ph, &pl, rh, rl, rh, rl);
	dd_mul_s(&qh, &ql, ph, pl, 0.5);
	dd_add_dd(&sh, &sl, rh, rl, qh, ql);
	dd_mul_dd(&ph, &pl, ph, pl, rh, rl);
	i=0;
	dd_mul_dd(&th, &tl, ph, pl, inv_fact[i][0], inv_fact[i][1]);
	i=i+1;
	
	while((i<=17) && (fabs(th)>thresh)) {
		dd_add_dd(&sh, &sl, sh, sl, th, tl);
		dd_mul_dd(&ph, &pl, ph, pl, rh, rl);
		dd_mul_dd(&th, &tl, ph, pl, inv_fact[i][0], inv_fact[i][1]);
		i=i+1;
	}
	dd_add_dd(&sh, &sl, sh, sl, th, tl);
    
	dd_sqr(&vh, &vl, sh, sl);
	dd_add_dd(&sh, &sl, 2.0*sh, 2.0*sl, vh, vl);    //1
	dd_sqr(&vh, &vl, sh, sl);
	dd_add_dd(&sh, &sl, 2.0*sh, 2.0*sl, vh, vl);    //2
	dd_sqr(&vh, &vl, sh, sl);
	dd_add_dd(&sh, &sl, 2.0*sh, 2.0*sl, vh, vl);    //3
	dd_sqr(&vh, &vl, sh, sl);
	dd_add_dd(&sh, &sl, 2.0*sh, 2.0*sl, vh, vl);    //4
	dd_sqr(&vh, &vl, sh, sl);
	dd_add_dd(&sh, &sl, 2.0*sh, 2.0*sl, vh, vl);    //5
	dd_sqr(&vh, &vl, sh, sl);
	dd_add_dd(&sh, &sl, 2.0*sh, 2.0*sl, vh, vl);    //6
	dd_sqr(&vh, &vl, sh, sl);
	dd_add_dd(&sh, &sl, 2.0*sh, 2.0*sl, vh, vl);    //7
	dd_sqr(&vh, &vl, sh, sl);
	dd_add_dd(&sh, &sl, 2.0*sh, 2.0*sl, vh, vl);    //8
	dd_sqr(&vh, &vl, sh, sl);
	dd_add_dd(&sh, &sl, 2.0*sh, 2.0*sl, vh, vl);    //9
	s_add_dd(&sh, &sl, 1.0, sh, sl);

	for(i=0; i<m; i++) {
		t=t*2;
	}

	dd_mul_s(&ehi[0], &elo[0], sh, sl, t);
	return(0);
}

inline void dd_add_dd(double *chi, double *clo, double ahi, double alo, double bhi, double blo)
{
	double sh,sl,eh,v;
	chi[0] = 0.0; clo[0] = 0.0;

	sh = ahi + bhi;
	v = sh - ahi;
	eh = (ahi - (sh - v)) + (bhi - v);
	
	sl = alo + blo;
	eh = eh + sl;
	chi[0] = sh + eh;
	clo[0] = eh - (chi[0] - sh);
	return;
}

inline void s_add_dd(double *chi, double *clo, double a, double bhi, double blo)
{
	double sh,eh,v;
		chi[0] = 0.0;
		clo[0] = 0.0;
		sh = a + bhi;
		v = sh - a;
		eh = (a - (sh - v)) + (bhi - v);
		eh = eh + blo;
		chi[0] = sh + eh;
		clo[0] = eh - (chi[0] - sh);
	
	return;
}

inline void dd_sub_dd(double *chi, double *clo, double ahi, double alo, double bhi, double blo)
{
	double sh,sl,eh,v;
	chi[0] = 0.0; clo[0] = 0.0;
	bhi = -bhi;	blo = -blo;
	
	sh = ahi + bhi;
	v = sh - ahi;
	eh = (ahi - (sh - v)) + (bhi - v);
	sl = alo + blo;
	eh = eh + sl;
	chi[0] = sh + eh;
	clo[0] = eh - (chi[0] - sh);
	return;
}

inline void s_sub_dd(double *chi, double *clo, double a, double bhi, double blo)
{
	double sh,eh,v;
		chi[0] = 0.0;
		clo[0] = 0.0;
		bhi = -bhi;
		blo = -blo;
		sh = a + bhi;
		v = sh - a;
		eh = (a - (sh - v)) + (bhi - v);
		eh = eh + blo;
		chi[0]= sh + eh;
		clo[0] = eh - (chi[0] - sh);
	
	return;
}


inline void dd_mul_dd(double *chi, double *clo, double ahi, double alo, double bhi, double blo)
{
	double t,ah,al,bh,bl,p1,p2;
	chi[0] = 0.0;	clo[0] = 0.0;

	t = 134217729 * ahi;
	ah = t - (t - ahi);
	al = ahi - ah;
	t = 134217729 * bhi;
	bh = t - (t - bhi);
	bl = bhi - bh;
	
	p1 = ahi * bhi;
	p2 = ((ah * bh - p1) + ah * bl + al * bh) + al * bl;
	p2 = p2 + ahi * blo;
	p2 = p2 + alo * bhi;
	
	chi[0] = p1 + p2;
	clo[0] = p2 - (chi[0] - p1);
	return;
}


inline void dd_mul_s(double *chi, double *clo, double ahi, double alo, double b)
{
	double t,ah,al,bh,bl,p1,p2,chi_n,clo_n;
	
	chi[0] = 0.0; clo[0] = 0.0;
	
	t = 134217729 * ahi;
	ah = t - (t - ahi);
	al = ahi - ah;
	
	t = 134217729 * b;
	bh = t - (t - b);
	bl = b - bh;
				
	p1 = ahi * b;
	p2 = ((ah * bh - p1) + ah * bl + al * bh) + al * bl;
    p2 = p2 + alo * b;
				
	chi[0] = p1 + p2;
	clo[0] = p2 - (chi[0] - p1);
	return;
}

inline void s_mul_dd(double *chi, double *clo, double a, double bhi, double blo)
{
	double p1,p2,t,ah,al,bh,bl;

	chi[0] = 0.0;	clo[0] = 0.0;

	t = 134217729 * a;
    ah = t - (t - a);
    al = a - ah;
	
    t = 134217729 * bhi;
    bh = t - (t - bhi);
    bl = bhi - bh;
	
	p1 = a * bhi;
	p2 = ((ah * bh - p1) + ah * bl + al * bh) + al * bl;
	p2 = p2 + a * blo;
	
	chi[0] = p1 + p2;
	clo[0] = p2 - (chi[0] - p1);
}

inline void dd_sqr(double *chi, double *clo, double ahi, double alo)
{
	double t,ah,al,p1,p2;
	chi[0] = 0.0;	clo[0] = 0.0;
    
	t = 134217729 * ahi;
	ah = t - (t - ahi);
	al = ahi - ah;
	
	p1 = ahi * ahi;
	p2 = ((ah * ah - p1) + 2.0 * (ah * al)) + al * al;
	p2 = p2 + 2.0 * (ahi * alo);
	
	chi[0] = p1 + p2;
	clo[0] = p2 - (chi[0] - p1);
	return;
}

inline void dd_div_dd(double *chi, double *clo, double ahi, double alo, double bhi, double blo)
{
	double c,cc,ah,al,bh,bl,p1,p2;
	chi[0] = 0.0; clo[0] = 0.0;
	
	c = ahi/bhi;
	p1 = c*bhi;
	cc = 134217729 * c;
	ah = cc - (cc - c);
	al = c - ah;
	cc = 134217729 * bhi;
	bh = cc - (cc - bhi);
	bl = bhi - bh;
	p2 = ((ah * bh - p1) + ah * bl + al * bh) + al * bl;
	
	cc = (ahi - p1 - p2 + alo - c*blo) / bhi;
	
	chi[0] = c + cc;
	clo[0] = cc - (chi[0] - c);
	return;
}

inline void ddsqrt_scl(double *chi, double *clo, double ahi, double alo)
{
	double c,cc,ah,al,p1,p2;
	
	chi[0] = 0.0;
	clo[0] = 0.0;
	
	c = sqrt(ahi);
	p1 = c*c;
	cc = 134217729 * c;
	ah = cc - (cc - c);
	al = c - ah;
	p2 = ((ah * ah - p1) + 2.0 * ah * al) + al * al;
	
	cc = (((ahi - p1) - p2) + alo) / (2.0 * c);
	
	chi[0] = c + cc;
	clo[0] = cc - (chi[0] - c);
	return;
}

inline void sin_taylor(double *chi, double *clo, double ahi, double alo)
{
	double eps, thresh, yhi, ylo, rhi, rlo, thi, tlo;
	int i=0;
	
	double fact[15][2] = {
	{1.66666666666666657e-01,  9.25185853854297066e-18},
	{4.16666666666666644e-02,  2.31296463463574266e-18},
	{8.33333333333333322e-03,  1.15648231731787138e-19},
	{1.38888888888888894e-03, -5.30054395437357706e-20},
	{1.98412698412698413e-04,  1.72095582934207053e-22},
	{2.48015873015873016e-05,  2.15119478667758816e-23},
	{2.75573192239858925e-06, -1.85839327404647208e-22},
	{2.75573192239858883e-07,  2.37677146222502973e-23},
	{2.50521083854417202e-08, -1.44881407093591197e-24},
	{2.08767569878681002e-09, -1.20834505911325997e-25},
	{1.60590438368216133e-10,  1.25852945887520981e-26},
	{1.14707455977297245e-11,  2.06555127528307454e-28},
	{7.64716373181981641e-13,  7.03872877733453001e-30},
	{4.77947733238738525e-14,  4.39920548583408126e-31},
	{2.81145725434552060e-15,  1.65088427308614326e-31}};
	
	eps = 4.93038065763132e-32;
	thresh = 0.5 * fabs(ahi) * eps; //!
	
	if((ahi==0)&&(alo==0)) {
		chi[0]=0.0;
		clo[0]=0.0;
		return;
	}
	
	dd_mul_dd(&yhi, &ylo, -ahi, -alo, ahi, alo);
	chi[0] = ahi;
	clo[0] = alo;
	rhi = ahi;
	rlo = alo;
	
	dd_mul_dd(&rhi, &rlo, rhi, rlo, yhi, ylo);
	dd_mul_dd(&thi, &tlo, rhi, rlo, fact[i][0], fact[i][1]);
	dd_add_dd(&chi[0], &clo[0], chi[0], clo[0], thi, tlo);
	i=i+2;
	
	while((i<15) && (fabs(thi)>thresh)) {
		dd_mul_dd(&rhi, &rlo, rhi, rlo, yhi, ylo);
		dd_mul_dd(&thi, &tlo, rhi, rlo, fact[i][0], fact[i][1]);
		dd_add_dd(&chi[0], &clo[0], chi[0], clo[0], thi, tlo);
		i=i+2;
	}
	
	return;
}

inline void cos_taylor(double *chi, double *clo, double ahi, double alo)
{
	double eps, thresh, yhi, ylo, rhi, rlo, thi, tlo, c=1.0,d=0.5;
	int i=1;
	
	double fact[15][2] = {
	{1.66666666666666657e-01,  9.25185853854297066e-18},
	{4.16666666666666644e-02,  2.31296463463574266e-18},
	{8.33333333333333322e-03,  1.15648231731787138e-19},
	{1.38888888888888894e-03, -5.30054395437357706e-20},
	{1.98412698412698413e-04,  1.72095582934207053e-22},
	{2.48015873015873016e-05,  2.15119478667758816e-23},
	{2.75573192239858925e-06, -1.85839327404647208e-22},
	{2.75573192239858883e-07,  2.37677146222502973e-23},
	{2.50521083854417202e-08, -1.44881407093591197e-24},
	{2.08767569878681002e-09, -1.20834505911325997e-25},
	{1.60590438368216133e-10,  1.25852945887520981e-26},
	{1.14707455977297245e-11,  2.06555127528307454e-28},
	{7.64716373181981641e-13,  7.03872877733453001e-30},
	{4.77947733238738525e-14,  4.39920548583408126e-31},
	{2.81145725434552060e-15,  1.65088427308614326e-31}};
	
	eps = 4.93038065763132e-32;	//2^-104
	thresh = 0.5 * eps; //!
	
	if((ahi==0)&&(alo==0)) {
		chi[0]=1.0;
		clo[0]=0.0;
		return;
	}
	
	dd_mul_dd(&yhi, &ylo, -ahi, -alo, ahi, alo);
	rhi = yhi;
	rlo = ylo;
	rhi = rhi/2;
	rlo = rlo/2;	
	s_add_dd(&chi[0], &clo[0], 1.0, rhi, rlo);
	rhi = rhi * 2;
	rlo = rlo * 2; 	

	dd_mul_dd(&rhi, &rlo, rhi, rlo, yhi, ylo);
	dd_mul_dd(&thi, &tlo, rhi, rlo, fact[i][0], fact[i][1]);
	dd_add_dd(&chi[0], &clo[0], chi[0], clo[0], thi, tlo);
	i=i+2;
	
	while((i<15) && (fabs(thi)>thresh)) {
		dd_mul_dd(&rhi, &rlo, rhi, rlo, yhi, ylo);
		dd_mul_dd(&thi, &tlo, rhi, rlo, fact[i][0], fact[i][1]);
		dd_add_dd(&chi[0], &clo[0], chi[0], clo[0], thi, tlo);
		i=i+2;
	}
	return;
}	

inline void sincos_taylor(double *shi, double *slo, double *chi, double *clo, double ahi, double alo)
{
	double bh, bl;
	double s=1.0;

	if((ahi==0) && (alo==0)) {
		shi[0]=0.0;
		slo[0]=0.0;
		chi[0]=1.0;
		clo[0]=0.0;
		return;
	}
	
	sin_taylor(&shi[0], &slo[0], ahi, alo) ;
	dd_mul_dd(&bh, &bl, shi[0], slo[0], shi[0], slo[0]);
	s_sub_dd(&bh, &bl, s, bh, bl);
	ddsqrt_scl(&chi[0], &clo[0], bh, bl);

	return;
}

inline void nint(double *hi, double *lo, double ahi, double alo)
{
	
	hi[0]=round(ahi);
	lo[0]=0.0;
	if(hi[0] == ahi) {
		lo[0] = round(alo);
		hi[0] = ahi+lo[0];
		lo[0] = lo[0] -(hi[0]-ahi);
	}
	else {
		lo[0]=0.0;
		if((fabs(hi[0]-ahi)==0.5) && (alo<0.0)) {
			hi[0] = hi[0]- 1.0;
		}
	}
	return;
}
