#include <math.h> 
inline void dd_add_dd(double *chi, double *clo, double ahi, double alo, double bhi, double blo);
inline void dd_mul_dd(double *chi, double *clo, double ahi, double alo, double bhi, double blo);
inline void dd_sub_dd(double *chi, double *clo, double ahi, double alo, double bhi, double blo);
inline void dd_div_dd(double *chi, double *clo, double ahi, double alo, double bhi, double blo);
inline void ddsqrt_scl(double *chi, double *clo, double ahi, double alo);

//******************************************************************************************//
// including files
//   ddgauss    : gaussian elimination
//   ddlu       : LU decomposition
//   ddinv      : inverse matrix
//   ddip       : inner product
//   ddnorm_two : vector 2-norm
//   ddsum      : vector summation
//   basic arithmetic for dd
//
// last modified : 2012/02/18
//
//******************************************************************************************//

// Gaussian elimination with pivoting
// For solving systems of linear equations with DD arithmetic
int ddgauss(double *xhi, double *xlo, double *ahi, double *alo, double *bhi, double *blo, int *n)
{
	int i,j,k,max_i;
	double c,cc,chi,clo,s,e,max,d,dd;
	//double *yhi,*ylo;
	//yhi = (double *)malloc(sizeof(double) * (*n));
	//ylo = (double *)malloc(sizeof(double) * (*n));

	// initialize 
	for(i=0;i<(*n);i++) {
		xhi[i] = 0;
		xlo[i] = 0;
	}
	
	//forward elimination
	for(k=0; k<(*n)-1; k++) {
		
		//pivotting
		max = fabs(ahi[(*n)*k+k]);
		max_i = k;
		for(i=k+1;i<(*n);i++) {
			if(fabs(ahi[(*n)*k+i]) > max) {
				max = fabs(ahi[(*n)*k+i]);
				max_i = i;
			}
		}

		//swap
		for(j=0;j<(*n);j++) {
			c = ahi[(*n)*j+k];
			cc= alo[(*n)*j+k];
			ahi[(*n)*j+k] = ahi[(*n)*j+max_i];
			alo[(*n)*j+k] = alo[(*n)*j+max_i];
			ahi[(*n)*j+max_i] = c;
			alo[(*n)*j+max_i] = cc;
		}
		c = bhi[k];
		cc= blo[k];
		bhi[k] = bhi[max_i];
		blo[k] = blo[max_i];
		bhi[max_i] = c;
		blo[max_i] = cc;

		for(i=k+1; i<(*n); i++) {
			c = 0; cc = 0;
			//d = a[i][k] / a[k][k];
			//division
			dd_div_dd(&d, &dd, ahi[(*n)*k+i], alo[(*n)*k+i], ahi[(*n)*k+k], alo[(*n)*k+k]);
			
			for(j=k+1; j<(*n); j++) {
				c = 0; cc = 0;
				//a[i][j] = a[i][j] - ( d * a[k][j] );
				//*[c,cc] = a[i][k] * a[k][j]
				//multiplication
				dd_mul_dd(&c, &cc, d, dd, ahi[(*n)*j+k], alo[(*n)*j+k]);
				//*a[i][j] = a[i][j] - [c,cc];
				//subtraction
				dd_sub_dd(&s, &e, ahi[(*n)*j+i], alo[(*n)*j+i], c, cc);
				ahi[(*n)*j+i] = s; alo[(*n)*j+i] = e;
			}
			c = 0;	cc = 0;
			//b[i] = b[i] - ( d * b[k] );
			dd_mul_dd(&c, &cc, d, dd, bhi[k], blo[k]);
			dd_sub_dd(&s, &e, bhi[i], blo[i], c, cc);
			bhi[i] = s;	blo[i] = e;
		}
	}

	//backward substitution
	for(i=(*n)-1; i>=0; i--) {
		c = bhi[i]; cc = blo[i];
		for(j=i+1; j<(*n); j++) {
			//[c,cc] -= a[i][j] * x[j];
			//multiplication
			dd_mul_dd(&chi, &clo, ahi[(*n)*j+i], alo[(*n)*j+i], xhi[j], xlo[j]);
			//subtraction
			dd_sub_dd(&s, &e, c, cc, chi, clo);
			c = s; cc= e;
		}
		//x[i] = (1.0 / A[i][i])*([c,cc]);
		//division
		dd_div_dd(&s, &e, c, cc, ahi[(*n)*i+i], alo[(*n)*i+i]);
		xhi[i] = s; xlo[i] = e;
	}
	
	//free(yhi);
	//free(ylo);
	return(0);
}


// LU decomposition
int ddlu(double *lhi, double *llo, double *uhi, double *ulo, double *ahi, double *alo, int *n)
{
	int i,j,k;
	double c,cc,s,e;

	// initialize 
	for(i=0;i<(*n);i++) {
		for(j=0;j<(*n);j++) {
			if(i == j) {
				lhi[(*n)*j+i] = 1.0; llo[(*n)*j+i] = 0.0;
				uhi[(*n)*j+i] = 0.0; ulo[(*n)*j+i] = 0.0;
			} else {
				lhi[(*n)*j+i] = 0.0; llo[(*n)*j+i] = 0.0;
				uhi[(*n)*j+i] = 0.0; ulo[(*n)*j+i] = 0.0;
			}
		}
		uhi[(*n)*i] = ahi[(*n)*i]; ulo[(*n)*i] = alo[(*n)*i];
	}
	
	for(k=0; k<(*n)-1; k++) {
		for(i=k+1; i<(*n); i++) {
			c = 0; cc = 0;
			//l[i][k] = a[i][k] / a[k][k];
			//division
			dd_div_dd(&s, &e, ahi[(*n)*k+i], alo[(*n)*k+i], ahi[(*n)*k+k], alo[(*n)*k+k]);
			lhi[(*n)*k+i] = s; llo[(*n)*k+i] = e;
			
			for(j=k; j<(*n); j++) {
				c = 0; cc = 0;
				//a[i][j] = a[i][j] - ( l[i][k] * a[k][j] );
				//*[c,cc] = l[i][k] * a[k][j]
				//multiplication
				dd_mul_dd(&c, &cc, lhi[(*n)*k+i], llo[(*n)*k+i], ahi[(*n)*j+k], alo[(*n)*j+k]);
				//*a[i][j] = a[i][j] - [c,cc];
				//subtraction			
				dd_sub_dd(&s, &e, ahi[(*n)*j+i], alo[(*n)*j+i], c, cc);
				ahi[(*n)*j+i] = s; alo[(*n)*j+i] = e;
				uhi[(*n)*j+i] = s; ulo[(*n)*j+i] = e;
			}
		}
	}
	return(0);
}


// compute inverse matrix of a
int ddinv(double *xhi, double *xlo, double *ahi, double *alo, double *yhi, double *ylo, int *n)
{
	int i,j,k;
	double c,cc,chi,clo,s,e;
	//double *yhi,*ylo;
	//yhi = (double *)malloc(sizeof(double) * (*n));
	//ylo = (double *)malloc(sizeof(double) * (*n));

	// initialize 
	for(i=0;i<(*n);i++) {
		for(j=0;j<(*n);j++) {
			if(i == j) {
				xhi[(*n)*j+i] = 1.0; xlo[(*n)*j+i] = 0.0;
			} else {
				xhi[(*n)*j+i] = 0.0; xlo[(*n)*j+i] = 0.0;
			}
		}
		yhi[i] = 0; ylo[i] = 0;
	}
	
	// LU decomposition
	for(k=0; k<(*n)-1; k++) {
		for(i=k+1; i<(*n); i++) {
			c = 0; cc = 0;
			//a[i][k] = a[i][k] / a[k][k];
			//division
			dd_div_dd(&s, &e, ahi[(*n)*k+i], alo[(*n)*k+i], ahi[(*n)*k+k], alo[(*n)*k+k]);
			ahi[(*n)*k+i] = s; alo[(*n)*k+i] = e;
			
			for(j=k+1; j<(*n); j++) {
				c = 0; cc = 0;
				//a[i][j] = a[i][j] - ( a[i][k] * a[k][j] );
				//*[c,cc] = a[i][k] * a[k][j]
				//multiplication
				dd_mul_dd(&c, &cc, ahi[(*n)*k+i], alo[(*n)*k+i], ahi[(*n)*j+k], alo[(*n)*j+k]);
				//*a[i][j] = a[i][j] - [c,cc];
				//subtraction			
				dd_sub_dd(&s, &e, ahi[(*n)*j+i], alo[(*n)*j+i], c, cc);
				ahi[(*n)*j+i] = s; alo[(*n)*j+i] = e;
			}
		}
	}

	for(i=0; i<(*n); i++) {
		yhi[0] = xhi[(*n)*i];
		ylo[0] = xlo[(*n)*i];
		for(k=1; k<(*n); k++) {
			c = 0; cc = 0;
			for(j=0; j<k; j++) {
				//[c,cc] += a[k][j] * y[j];
				//multiplication
				dd_mul_dd(&chi, &clo, ahi[(*n)*j+k], alo[(*n)*j+k], yhi[j], ylo[j]);
				//addition
				dd_add_dd(&s, &e, c, cc, chi, clo);
				c = s; cc= e;
			}
			//y[k] = x[k][i] - [c,cc];
			//subtraction
			dd_sub_dd(&s, &e, xhi[(*n)*i+k], xlo[(*n)*i+k], c, cc);
			yhi[k] = s; ylo[k] = e;
		}

		for(k=(*n)-1; k>=0; k--) {
			c = 0; cc = 0;
			for(j=k+1; j<(*n); j++) {
				//[c,cc] += a[k][j] * x[j][i];
				//multiplication
				dd_mul_dd(&chi, &clo, ahi[(*n)*j+k], alo[(*n)*j+k], xhi[(*n)*i+j], xlo[(*n)*i+j]);

				//addition
				dd_add_dd(&s, &e, c, cc, chi, clo);
				c = s; cc= e;
			}
			//x[k][i] = (1.0 / A[k][k])*(y[k] - [c,cc]);
			//subtraction
			dd_sub_dd(&s, &e, yhi[k], ylo[k], c, cc);
			xhi[(*n)*i+k] = s; xlo[(*n)*i+k] = e;
			//division
			dd_div_dd(&s, &e, xhi[(*n)*i+k], xlo[(*n)*i+k], ahi[(*n)*k+k], alo[(*n)*k+k]);
			xhi[(*n)*i+k] = s; xlo[(*n)*i+k] = e;
		}
	}
	//free(yhi);
	//free(ylo);
	return(0);
}

int ddip(double *c, double *xhi, double *xlo, double *yhi, double *ylo, int *m)
{
	c[0] = 0.0;
	c[1] = 0.0;
	int i;
	double chi,clo,s,e;
	chi = 0; clo = 0; s = 0; e = 0;
    
	for (i = 0 ; i < *m ; i++) {
		//multiplication
		dd_mul_dd(&chi, &clo, xhi[i], xlo[i], yhi[i], ylo[i]);
		//addition
		dd_add_dd(&s,&e,c[0],c[1],chi,clo);
		c[0] = s;
		c[1] = e;
	}
	return(0);
}

int ddnorm_two(double *c, double *xhi, double *xlo, int *m)
{
	c[0] = 0.0;
	c[1] = 0.0;
	int i;
	double chi,clo,s,e;
	s = 0; e = 0;
    
	dd_mul_dd(&s, &e, xhi[0], xlo[0], xhi[0], xlo[0]);
	c[0] = s; c[1] = e;
	
	for (i = 1 ; i < *m ; i++) {
		dd_mul_dd(&chi, &clo, xhi[i], xlo[i], xhi[i], xlo[i]);
		dd_add_dd(&s,&e,c[0],c[1],chi,clo);
		c[0] = s; c[1] = e;
	}
	ddsqrt_scl(&s, &e, c[0], c[1]);
	c[0] = s; c[1] = e;
	return(0);
}


int ddsum(double *c, double *xhi, double *xlo, int *m, int *n)
{
	c[0] = 0.0;
	c[1] = 0.0;
	int i;
	double s,e;
	s = 0; e = 0;
    
	c[0] = xhi[0]; c[1] = xlo[0];
	
	for (i = 1 ; i < (*m)*(*n) ; i++) {
		dd_add_dd(&s,&e,c[0],c[1],xhi[i],xlo[i]);
		c[0] = s; c[1] = e;
	}
	return(0);
}




inline void dd_add_dd(double *chi, double *clo, double ahi, double alo, double bhi, double blo)
{
	double sh,sl,eh,v;
	chi[0] = 0.0; clo[0] = 0.0;

	sh = ahi + bhi;
	v = sh - ahi;
	eh = (ahi - (sh - v)) + (bhi - v);
	
	sl = alo + blo;
	eh = eh + sl;
	chi[0]= sh + eh;
	clo[0] = eh - (chi[0] - sh);
	return;
}

inline void dd_sub_dd(double *chi, double *clo, double ahi, double alo, double bhi, double blo)
{
	double sh,sl,eh,v;
	chi[0] = 0.0; clo[0] = 0.0;
	bhi = -bhi;	blo = -blo;
	
	sh = ahi + bhi;
	v = sh - ahi;
	eh = (ahi - (sh - v)) + (bhi - v);
	sl = alo + blo;
	eh = eh + sl;
	chi[0] = sh + eh;
	clo[0] = eh - (chi[0] - sh);
	return;
}


inline void dd_mul_dd(double *chi, double *clo, double ahi, double alo, double bhi, double blo)
{
	double t,ah,al,bh,bl,p1,p2;
	chi[0] = 0.0;	clo[0] = 0.0;

	t = 134217729 * ahi;
	ah = t - (t - ahi);
	al = ahi - ah;
	t = 134217729 * bhi;
	bh = t - (t - bhi);
	bl = bhi - bh;
	
	p1 = ahi * bhi;
	p2 = ((ah * bh - p1) + ah * bl + al * bh) + al * bl;
	p2 = p2 + ahi * blo;
	p2 = p2 + alo * bhi;
	
	chi[0] = p1 + p2;
	clo[0] = p2 - (chi[0] - p1);
	return;
}

inline void dd_div_dd(double *chi, double *clo, double ahi, double alo, double bhi, double blo)
{
	double c,cc,ah,al,bh,bl,p1,p2;
	chi[0] = 0.0; clo[0] = 0.0;
	
	c = ahi/bhi;
	p1 = c*bhi;
	cc = 134217729 * c;
	ah = cc - (cc - c);
	al = c - ah;
	cc = 134217729 * bhi;
	bh = cc - (cc - bhi);
	bl = bhi - bh;
	p2 = ((ah * bh - p1) + ah * bl + al * bh) + al * bl;
	
	cc = (ahi - p1 - p2 + alo - c*blo) / bhi;
	
	chi[0] = c + cc;
	clo[0] = cc - (chi[0] - c);
	return;
}

inline void ddsqrt_scl(double *chi, double *clo, double ahi, double alo)
{
	double c,cc,ah,al,p1,p2;
	
	chi[0] = 0.0;
	clo[0] = 0.0;
	
	c = sqrt(ahi);
	p1 = c*c;
	cc = 134217729 * c;
	ah = cc - (cc - c);
	al = c - ah;
	p2 = ((ah * ah - p1) + 2.0 * ah * al) + al * al;
	
	cc = (((ahi - p1) - p2) + alo) / (2.0 * c);
	
	chi[0] = c + cc;
	clo[0] = cc - (chi[0] - c);
	return;
}
