#include <stdbool.h>
#include <math.h>

inline void fast_two_sum(double *s, double *e, double a, double b);
inline void two_sum(double *s, double *e, double a, double b);
inline void three_sum(double *d, double *e, double *f, double a, double b, double c);
inline void three_sum2(double *d, double *e, double a, double b, double c);
inline void two_prod(double *p, double *e, double a, double b);
inline void fast_three_accum(double *s, double *t, double *u, double a, double b, double c);
inline void renormalize(double *b0, double *b1, double *b2, double *b3, double a0, double a1, double a2, double a3, double a4);
inline void qd_add_s(double *c0, double *c1, double *c2, double *c3, double a0, double a1, double a2, double a3, double b);
inline void qd_add_dd(double *c0, double *c1, double *c2, double *c3, double a0, double a1, double a2, double a3, double b0, double b1);
inline void qd_add_qd(double *c0, double *c1, double *c2, double *c3, double a0, double a1, double a2, double a3, double b0, double b1, double b2, double b3);
inline void s_mul_dd(double *chi, double *clo, double a, double bhi, double blo);
inline void s_mul_qd(double *c0, double *c1, double *c2, double *c3, double b, double a0, double a1, double a2, double a3);
inline void dd_mul_qd(double *c0, double *c1, double *c2, double *c3, double b0, double b1, double a0, double a1, double a2, double a3);
inline void qd_mul_dd(double *c0, double *c1, double *c2, double *c3, double a0, double a1, double a2, double a3, double b0, double b1);
inline void qd_mul_qd(double *c0, double *c1, double *c2, double *c3, double a0, double a1, double a2, double a3, double b0, double b1, double b2, double b3);


//******************************************************************************************//
// four basic arithmetic of quad-double 
//  - addition
//  - subtraction
//  - multiplication
//  - division
// renormalization
// error-free transformation
//
// last modified : 2012/02/03
//
//******************************************************************************************//


//--------------------------------------------------------------------------------------------
//
// quad-double addition
//
// args
// a0, a1, a2, a3, b0, b1, b2, b3 : double numbers
// a0 + a1 = dd nubmer, a0 + a1 + a2 + a3 = qd number
// *m : row
// *n : column
// return (c0,c1,c2,c3) for qd number
//
// routines
// qd_a_s  : qd - double addition
// s_a_qd  : double - qd addition
// qd_a_dd : qd - dd addition
// dd_a_qd : dd - qd addition
// qd_a_qd : qd - qd addition (ieee add)
//
//--------------------------------------------------------------------------------------------


//quad-double addition
//qd + dobule
int qd_a_s(double *c0, double *c1, double *c2, double *c3, double *a0, double *a1, double *a2, double *a3, double *b, int *m, int *n)
{
	double e,x,y,w,z;
	e=0.0;
	int i;
	for(i=0; i<(*m)*(*n); i++) {
		c0[i] = 0.0; c1[i] = 0.0; c2[i] = 0.0; c3[i] = 0.0;
        
        two_sum(&x, &y, a0[i], b[i]);
		c0[i] = x;
		e = y;
		two_sum(&x, &y, a1[i], e);
		c1[i] = x;
		e = y;
		two_sum(&x, &y, a2[i], e);
		c2[i] = x;
		e = y;
		two_sum(&x, &y, a3[i], e);
		c3[i] = x;
		e = y;
		renormalize(&x, &y, &w, &z, c0[i], c1[i], c2[i], c3[i], e);
		c0[i] = x;
		c1[i] = y;
		c2[i] = w;
		c3[i] = z;
        
		/*two_sum(&c0[i], &e, a0[i], b[i]);
		two_sum(&c1[i], &e, a1[i], e);
		two_sum(&c2[i], &e, a2[i], e);
		two_sum(&c3[i], &e, a3[i], e);
		renormalize(&c0[i], &c1[i], &c2[i], &c3[i], c0[i], c1[i], c2[i], c3[i], e);*///でも大丈夫
	}
  return(0);
}

//quad-double addition
//double + qd
int s_a_qd(double *c0, double *c1, double *c2, double *c3, double *a, double *b0, double *b1, double *b2, double *b3, int *m, int *n)
{
	double e,x,y,w,z;
	e=0.0;
	int i;
	for(i=0; i<(*m)*(*n); i++) {
		c0[i] = 0.0; c1[i] = 0.0; c2[i] = 0.0; c3[i] = 0.0;

		two_sum(&x, &y, a[i], b0[i]);
		c0[i] = x;
		e = y;
		two_sum(&x, &y, b1[i], e);
		c1[i] = x;
		e = y;
		two_sum(&x, &y, b2[i], e);
		c2[i] = x;
		e = y;
		two_sum(&x, &y, b3[i], e);
		c3[i] = x;
		e = y;
		renormalize(&x, &y, &w, &z, c0[i], c1[i], c2[i], c3[i], e);
		c0[i] = x;
		c1[i] = y;
		c2[i] = w;
		c3[i] = z;
	}
  return(0);
}

//quad-double addition
//qd + dd
int qd_a_dd(double *c0, double *c1, double *c2, double *c3, double *a0, double *a1, double *a2, double *a3, double *b0, double *b1, int *m, int *n)
{
	double e1,e2,x,y,w,z;
	e1 = 0.0; e2 = 0.0;
	int i;

	for(i=0; i<(*m)*(*n); i++) {
		c0[i] = 0.0; c1[i] = 0.0; c2[i] = 0.0; c3[i] = 0.0;

		two_sum(&x, &y, a0[i], b0[i]);
		c0[i] = x;
		e1 = y;

		two_sum(&x, &y, a1[i], b1[i]);
		c1[i] = x;
		e2 = y;

		two_sum(&x, &y, c1[i], e1);
		c1[i] = x;
		e1 = y;
  
		two_sum(&x, &y, a2[i], e2);
		c2[i] = x;
		e2 = y;

  		two_sum(&x, &y, c2[i], e1);
		c2[i] = x;
		e1 = y;

		two_sum(&x, &y, e1, e2);
		e1 = x;
		e2 = y;

		two_sum(&x, &y, a3[i], e1);
		c3[i] = x;
		e1 = y;

		e1 = e1 + e2;

		renormalize(&x, &y, &w, &z, c0[i], c1[i], c2[i], c3[i], e1);
		c0[i] = x;
		c1[i] = y;
		c2[i] = w;
		c3[i] = z;
	}
  return(0);
}

//qd addition
//dd + qd
int dd_a_qd(double *c0, double *c1, double *c2, double *c3, double *a0, double *a1, double *b0, double *b1, double *b2, double *b3, int *m, int *n)
{
	double e1,e2,x,y,w,z;
	e1 = 0.0; e2 = 0.0;
	int i;
    
	for(i=0; i<(*m)*(*n); i++) {
		c0[i] = 0.0; c1[i] = 0.0; c2[i] = 0.0; c3[i] = 0.0;
        
		two_sum(&x, &y, a0[i], b0[i]);
		c0[i] = x;
		e1 = y;
        
		two_sum(&x, &y, a1[i], b1[i]);
		c1[i] = x;
		e2 = y;
        
		two_sum(&x, &y, c1[i], e1);
		c1[i] = x;
		e1 = y;
        
		two_sum(&x, &y, b2[i], e2);
		c2[i] = x;
		e2 = y;
        
  		two_sum(&x, &y, c2[i], e1);
		c2[i] = x;
		e1 = y;
        
		two_sum(&x, &y, e1, e2);
		e1 = x;
		e2 = y;
        
		two_sum(&x, &y, b3[i], e1);
		c3[i] = x;
		e1 = y;
        
		e1 = e1 + e2;
        
		renormalize(&x, &y, &w, &z, c0[i], c1[i], c2[i], c3[i], e1);
		c0[i] = x;
		c1[i] = y;
		c2[i] = w;
		c3[i] = z;
	}
    return(0);
}

//qd addition
//qd + qd
/*int qd_a_qd(double *c0, double *c1, double *c2, double *c3, double *a0, double *a1, double *a2, double *a3, double *b0, double *b1, double *b2, double *b3,int *m, int *n)
{
	double t,u,v,w,x,y,z;
	int s,i,j,k;
    double c[4] = {0.0, 0.0, 0.0, 0.0};
    double a[4] = {0.0, 0.0, 0.0, 0.0};
    double b[4] = {0.0, 0.0, 0.0, 0.0};
    
	for(s=0; s<(*m)*(*n); s++) {
		c0[s] = 0.0; c1[s] = 0.0; c2[s] = 0.0; c3[s] = 0.0;
        a[0] = a0[s]; a[1] = a1[s]; a[2] = a2[s]; a[3] = a3[s];
        b[0] = b0[s]; b[1] = b1[s]; b[2] = b2[s]; b[3] = b3[s];
        u = 0.0; v = 0.0;
        i = 0; j = 0; k = 0;

        if (fabs(a[i])>fabs(b[j])) { // a_0 > b_0
            u = a[i];
            i = i + 1;
        } else {                   // b_0 > a_0
            u = b[j];
            j = j + 1;
        }
        
        if (fabs(a[i])>fabs(b[j])) { // (a_0 > a_1 > b_0) or (b_0 > a_0 > b_1)
            v = a[i];
            i = i + 1;
        } else {                     // (a_0 > b_0 > a_1) or (b_0 > b_1 > a_0)
            v = b[j];
            j = j + 1;
        }
        // 1. u = a_0, v = a_1
        // 2. u = a_0, v = b_0
        // 3. u = b_0, v = a_0
        // 4. u = b_0, v = b_1
        
		fast_two_sum(&x, &y, u, v);
		u = x;
		v = y;
        
        while (k < 4) {
            //if all component was substituted
            if (i >= 4 && j >= 4) {
                c[k] = u;
                if (k < 3) {
                    k = k + 1;
                    c[k] = v;
                }
                break;
            }
            
            if (i >= 4) { // if all a's components are substituted
                t = b[j]; // substitute the next b
                j = j + 1;
            } else if (j >= 4) { // if all b's components are substituted
                t = a[i]; // substitute the next a
                i = i + 1;
            } else if (fabs(a[i]) > fabs(b[j])) {
                t = a[i]; // if a[i] > b[j] then substitute the next a
                i = i + 1;
            } else { 
                t = b[j]; // if a[i] < b[j] then substitute the next b
                j = j + 1;
            }
            
            fast_three_accum(&x, &y, &z, u, v, t);
            u = y;
            v = z;
            if(x != 0.0) {
                c[k] = x;
                k = k + 1;
            }
        }
        
        for (k = i; k < 4; k++) {
            c[3] = c[3] + a[k];
        }
        for (k = j; k < 4; k++) {
            c[3] = c[3] + b[k];
        }
        
		renormalize(&x, &y, &w, &z, c[0], c[1], c[2], c[3], 0);
		c0[s] = x;
		c1[s] = y;
		c2[s] = w;
		c3[s] = z;
	}
    return(0);
}*/

int qd_a_qd(double *c0, double *c1, double *c2, double *c3, double *a0, double *a1, double *a2, double *a3, double *b0, double *b1, double *b2, double *b3,int *m, int *n)
{
	double e1,e2,e3,e4,x,y,w,z;
	e1 = 0.0; e2 = 0.0;
	int i;
    
	for(i=0; i<(*m)*(*n); i++) {
		c0[i] = 0.0; c1[i] = 0.0; c2[i] = 0.0; c3[i] = 0.0;
        
		two_sum(&x, &y, a0[i], b0[i]);		
		c0[i] = x;
		e1 = y;
		
		two_sum(&x, &y, a1[i], b1[i]);
		c1[i] = x;
		e2 = y;
		
		two_sum(&x, &y, c1[i], e1);
		c1[i] = x;
		e1 = y;
        
		two_sum(&x, &y, a2[i], b2[i]);
		c2[i] = x;
		e3 = y;
        
		three_sum(&x, &y, &z, c2[i], e2, e1);
		c2[i] = x;
		e1 = y;
		e2 = z;
        
		two_sum(&x, &y, a3[i], b3[i]);
		c3[i] = x;
		e4 = y;
        
		three_sum2(&x, &y, c3[i], e3, e1);
		c3[i] = x;
		e1 = y;
        
		e1 = e1 + e2 + e4;
        
		renormalize(&x, &y, &w, &z, c0[i], c1[i], c2[i], c3[i], e1);
		c0[i] = x;
		c1[i] = y;
		c2[i] = w;
		c3[i] = z;
	}
	return(0);
}


//--------------------------------------------------------------------------------------------
//
// quad-double subtraction
//
// args
// a0, a1, a2, a3, b0, b1, b2, b3 : double numbers
// a0 + a1 = dd nubmer, a0 + a1 + a2 + a3 = qd number
// *m : row
// *n : column
// return (c0,c1,c2,c3) for qd number
//
// routines
// qd_s_s  : qd - double subtraction
// s_s_qd  : double - qd subtraction
// qd_s_dd : qd - dd subtraction
// dd_s_qd : dd - qd subtraction
// qd_s_qd : qd - qd subtraction
//
//--------------------------------------------------------------------------------------------

//qd subtruction
//qd - d
int qd_s_s(double *c0, double *c1, double *c2, double *c3, double *a0, double *a1, double *a2, double *a3, double *b, int *m, int *n)
{
	double e,x,y,w,z;
	e=0.0;
	int i;
	for(i=0; i<(*m)*(*n); i++) {
		c0[i] = 0.0; c1[i] = 0.0; c2[i] = 0.0; c3[i] = 0.0;
		b[i] = -b[i];

		two_sum(&x, &y, a0[i], b[i]);
		c0[i] = x;
		e = y;

		two_sum(&x, &y, a1[i], e);
		c1[i] = x;
		e = y;

		two_sum(&x, &y, a2[i], e);
		c2[i] = x;
		e = y;

		two_sum(&x, &y, a3[i], e);
		c3[i] = x;
		e = y;

		renormalize(&x, &y, &w, &z, c0[i], c1[i], c2[i], c3[i], e);
		c0[i] = x;
		c1[i] = y;
		c2[i] = w;
		c3[i] = z;
	}
  return(0);
}

//qd subtruction
//d - qd
int s_s_qd(double *c0, double *c1, double *c2, double *c3, double *a, double *b0, double *b1, double *b2, double *b3, int *m, int *n)
{
	double e,x,y,w,z;
	e=0.0;
	int i;
	for(i=0; i<(*m)*(*n); i++) {
		c0[i] = 0.0; c1[i] = 0.0; c2[i] = 0.0; c3[i] = 0.0;
		b0[i] = -b0[i];	b1[i] = -b1[i];	b2[i] = -b2[i];	b3[i] = -b3[i];

		two_sum(&x, &y, a[i], b0[i]);
		c0[i] = x;
		e = y;
		two_sum(&x, &y, b1[i], e);
		c1[i] = x;
		e = y;
		two_sum(&x, &y, b2[i], e);
		c2[i] = x;
		e = y;
		two_sum(&x, &y, b3[i], e);
		c3[i] = x;
		e = y;
		renormalize(&x, &y, &w, &z, c0[i], c1[i], c2[i], c3[i], e);
		c0[i] = x;
		c1[i] = y;
		c2[i] = w;
		c3[i] = z;
	}
  return(0);
}

//qd subtruction
//qd - dd
int qd_s_dd(double *c0, double *c1, double *c2, double *c3, double *a0, double *a1, double *a2, double *a3, double *b0, double *b1, int *m, int *n)
{
	double e1,e2,x,y,w,z;
	e1 = 0.0; e2 = 0.0;
	int i;

	for(i=0; i<(*m)*(*n); i++) {
		c0[i] = 0.0; c1[i] = 0.0; c2[i] = 0.0; c3[i] = 0.0;
		b0[i] = -b0[i];	b1[i] = -b1[i];

		two_sum(&x, &y, a0[i], b0[i]);
		c0[i] = x;
		e1 = y;

		two_sum(&x, &y, a1[i], b1[i]);
		c1[i] = x;
		e2 = y;

		two_sum(&x, &y, c1[i], e1);
		c1[i] = x;
		e1 = y;

  		two_sum(&x, &y, a2[i], e2);
		c2[i] = x;
		e2 = y;

		two_sum(&x, &y, c2[i], e1);
		c2[i] = x;
		e1 = y;

		two_sum(&x, &y, e1, e2);
		e1 = x;
		e2 = y;

		two_sum(&x, &y, a3[i], e1);
		c3[i] = x;
		e1 = y;

		e1 = e1 + e2;

		renormalize(&x, &y, &w, &z, c0[i], c1[i], c2[i], c3[i], e1);
		c0[i] = x;
		c1[i] = y;
		c2[i] = w;
		c3[i] = z;
	}
  return(0);
}

//qd subtruction
//dd - qd
int dd_s_qd(double *c0, double *c1, double *c2, double *c3, double *a0, double *a1, double *b0, double *b1, double *b2, double *b3, int *m, int *n)
{
	double e1,e2,x,y,w,z;
	e1 = 0.0; e2 = 0.0;
	int i;

	for(i=0; i<(*m)*(*n); i++) {
		c0[i] = 0.0; c1[i] = 0.0; c2[i] = 0.0; c3[i] = 0.0;
		b0[i] = -b0[i];	b1[i] = -b1[i];	b2[i] = -b2[i];	b3[i] = -b3[i];

		two_sum(&x, &y, a0[i], b0[i]);
		c0[i] = x;
		e1 = y;

		two_sum(&x, &y, a1[i], b1[i]);
		c1[i] = x;
		e2 = y;

		two_sum(&x, &y, c1[i], e1);
		c1[i] = x;
		e1 = y;
  
		two_sum(&x, &y, b2[i], e2);
		c2[i] = x;
		e2 = y;

  		two_sum(&x, &y, c2[i], e1);
		c2[i] = x;
		e1 = y;

		two_sum(&x, &y, e1, e2);
		e1 = x;
		e2 = y;

		two_sum(&x, &y, b3[i], e1);
		c3[i] = x;
		e1 = y;

		e1 = e1 + e2;

		renormalize(&x, &y, &w, &z, c0[i], c1[i], c2[i], c3[i], e1);
		c0[i] = x;
		c1[i] = y;
		c2[i] = w;
		c3[i] = z;
	}
  return(0);
}

//qd subtruction
//qd - qd
/*int qd_s_qd(double *c0, double *c1, double *c2, double *c3, double *a0, double *a1, double *a2, double *a3, double *b0, double *b1, double *b2, double *b3,int *m, int *n)
{
	double t,u,v,w,x,y,z;
	int s,i,j,k;
    double c[4] = {0.0, 0.0, 0.0, 0.0};
    double a[4] = {0.0, 0.0, 0.0, 0.0};
    double b[4] = {0.0, 0.0, 0.0, 0.0};
    
	for(s=0; s<(*m)*(*n); s++) {
		c0[s] = 0.0; c1[s] = 0.0; c2[s] = 0.0; c3[s] = 0.0;
        a[0] = a0[s]; a[1] = a1[s]; a[2] = a2[s]; a[3] = a3[s];
        b[0] = -b0[s]; b[1] = -b1[s]; b[2] = -b2[s]; b[3] = -b3[s];
        u = 0.0; v = 0.0;
        i = 0; j = 0; k = 0;
        
        if (fabs(a[i])>fabs(b[j])) { // a_0 > b_0
            u = a[i];
            i = i + 1;
        } else {                   // b_0 > a_0
            u = b[j];
            j = j + 1;
        }
        
        if (fabs(a[i])>fabs(b[j])) { // (a_0 > a_1 > b_0) or (b_0 > a_0 > b_1)
            v = a[i];
            i = i + 1;
        } else {                     // (a_0 > b_0 > a_1) or (b_0 > b_1 > a_0)
            v = b[j];
            j = j + 1;
        }
        // 1. u = a_0, v = a_1
        // 2. u = a_0, v = b_0
        // 3. u = b_0, v = a_0
        // 4. u = b_0, v = b_1
        
		fast_two_sum(&x, &y, u, v);
		u = x;
		v = y;
        
        while (k < 4) {
            //if all component was substituted
            if (i >= 4 && j >= 4) {
                c[k] = u;
                if (k < 3) {
                    k = k + 1;
                    c[k] = v;
                }
                break;
            }
            
            if (i >= 4) { // if all a's components are substituted
                t = b[j]; // substitute the next b
                j = j + 1;
            } else if (j >= 4) { // if all b's components are substituted
                t = a[i]; // substitute the next a
                i = i + 1;
            } else if (fabs(a[i]) > fabs(b[j])) {
                t = a[i]; // if a[i] > b[j] then substitute the next a
                i = i + 1;
            } else { 
                t = b[j]; // if a[i] < b[j] then substitute the next b
                j = j + 1;
            }
            
            fast_three_accum(&x, &y, &z, u, v, t);
            u = y;
            v = z;
            if(x != 0.0) {
                c[k] = x;
                k = k + 1;
            }
        }
        
        for (k = i; k < 4; k++) {
            c[3] = c[3] + a[k];
        }
        for (k = j; k < 4; k++) {
            c[3] = c[3] + b[k];
        }
        
		renormalize(&x, &y, &w, &z, c[0], c[1], c[2], c[3], 0);
		c0[s] = x;
		c1[s] = y;
		c2[s] = w;
		c3[s] = z;
	}
    return(0);
}*/
int qd_s_qd(double *c0, double *c1, double *c2, double *c3, double *a0, double *a1, double *a2, double *a3, double *b0, double *b1, double *b2, double *b3,int *m, int *n)
{
	double e1,e2,e3,e4,x,y,w,z;
	e1 = 0.0; e2 = 0.0;
	int i;
    
	for(i=0; i<(*m)*(*n); i++) {
		c0[i] = 0.0; c1[i] = 0.0; c2[i] = 0.0; c3[i] = 0.0;
		b0[i] = -b0[i];	b1[i] = -b1[i];	b2[i] = -b2[i];	b3[i] = -b3[i];
        
		two_sum(&x, &y, a0[i], b0[i]);		
		c0[i] = x;
		e1 = y;
		
		two_sum(&x, &y, a1[i], b1[i]);
		c1[i] = x;
		e2 = y;
		
		two_sum(&x, &y, c1[i], e1);
		c1[i] = x;
		e1 = y;
        
		two_sum(&x, &y, a2[i], b2[i]);
		c2[i] = x;
		e3 = y;
        
		three_sum(&x, &y, &z, c2[i], e2, e1);
		c2[i] = x;
		e1 = y;
		e2 = z;
        
		two_sum(&x, &y, a3[i], b3[i]);
		c3[i] = x;
		e4 = y;
        
		three_sum2(&x, &y, c3[i], e3, e1);
		c3[i] = x;
		e1 = y;
        
		e1 = e1 + e2 + e4;
        
		renormalize(&x, &y, &w, &z, c0[i], c1[i], c2[i], c3[i], e1);
		c0[i] = x;
		c1[i] = y;
		c2[i] = w;
		c3[i] = z;
	}
	return(0);
}

//--------------------------------------------------------------------------------------------
//
// quad-double multiplication
//
// args
// a0, a1, a2, a3, b0, b1, b2, b3 : double numbers
// a0 + a1 = dd nubmer, a0 + a1 + a2 + a3 = qd number
// *m : row
// *n : column
// return (c0,c1,c2,c3) for qd number
//
// routines
// qd_d_scalar_multi_1 : qd(scalar) * double(matrix)
// qd_d_scalar_multi_2 : qd(matrix) * double(scalar)
// qd_m_s  : qd - double multiplication
// s_s_qd  : double - qd multiplication
// qd_dd_scalar_multi_1 : qd(scalar) * dd(matrix)
// qd_dd_scalar_multi_2 : qd(matrix) * dd(scalar)
// qd_m_dd : qd - dd multiplication
// dd_m_qd : dd - qd multiplication
// qd_m_qd : qd - qd multiplication
// scalar_multi_qd : qd(scalar) * qd(matrix)
//
//--------------------------------------------------------------------------------------------

// qd_d_scalar_multi_1
// qd(scalar) * double(matrix)
int qd_d_scalar_multi_1(double *c0, double *c1, double *c2, double *c3, double *a0, double *a1, double *a2, double *a3, double *b, int *m, int *n)
{
	double e1,e2,e3,x,y,w,z;
	int i;

	for(i=0; i<(*m)*(*n); i++) {
		c0[i] = 0.0; c1[i] = 0.0; c2[i] = 0.0; c3[i] = 0.0;
		
		two_prod(&x, &y, a0[0], b[i]);
		c0[i] = x;
		e1 = y;

		two_prod(&x, &y, a1[0], b[i]);
		c1[i] = x;
		e2 = y;

		two_sum(&x, &y, c1[i], e1);
		c1[i] = x;
		e1 = y;

		two_prod(&x, &y, a2[0], b[i]);
		c2[i] = x;
		e3 = y;
		
		three_sum(&x, &y, &z, c2[i], e2, e1);
		c2[i] = x;
		e1 = y;
		e2 = z;

		c3[i] = a3[0]*b[i];
		
		three_sum2(&x, &y, c3[i], e3, e1);
		c3[i] = x;
		e1 = y;

		e1 = e1 + e2;
		
		renormalize(&x, &y, &w, &z, c0[i], c1[i], c2[i], c3[i], e1);
		c0[i] = x;
		c1[i] = y;
		c2[i] = w;
		c3[i] = z;
	}
	return(0);
}

// qd_d_scalar_multi_2
// qd(matrix) * double(scalar)
int qd_d_scalar_multi_2(double *c0, double *c1, double *c2, double *c3, double *a0, double *a1, double *a2, double *a3, double *b, int *m, int *n)
{
	double e1,e2,e3,x,y,w,z;
	int i;

	for(i=0; i<(*m)*(*n); i++) {
		c0[i] = 0.0; c1[i] = 0.0; c2[i] = 0.0; c3[i] = 0.0;
		
		two_prod(&x, &y, a0[i], b[0]);	c0[i] = x;	e1 = y;
		two_prod(&x, &y, a1[i], b[0]);	c1[i] = x;	e2 = y;
		two_sum(&x, &y, c1[i], e1);		c1[i] = x;	e1 = y;
		two_prod(&x, &y, a2[i], b[0]);	c2[i] = x;	e3 = y;
		
		three_sum(&x, &y, &z, c2[i], e2, e1);	c2[i] = x;	e1 = y;	e2 = z;

		c3[i] = a3[i]*b[0];
		
		three_sum2(&x, &y, c3[i], e3, e1);		c3[i] = x;	e1 = y;

		e1 = e1 + e2;
		
		renormalize(&x, &y, &w, &z, c0[i], c1[i], c2[i], c3[i], e1);
		c0[i] = x;	c1[i] = y;	c2[i] = w;	c3[i] = z;
	}
	return(0);
}

// qd multiplication
// qd * double
int qd_m_s(double *c0, double *c1, double *c2, double *c3, double *a0, double *a1, double *a2, double *a3, double *b, int *m, int *n, int *l)
{
	double p,q,r,s,x,y,w,z;
	int i,j,k;
	
	for (i = 0 ; i < *m ; i++) {
		for (j = 0 ; j < *n ; j++) {	
			c0[(*m)*j+i] = 0.0;
			c1[(*m)*j+i] = 0.0;
			c2[(*m)*j+i] = 0.0;
			c3[(*m)*j+i] = 0.0;
			for (k = 0 ; k < *l ; k++) {
				//multiplication
				s_mul_qd(&x, &y, &w, &z, b[(*l)*j+k], a0[(*m)*k+i], a1[(*m)*k+i], a2[(*m)*k+i], a3[(*m)*k+i]);
				//addition
				qd_add_qd(&p, &q, &r, &s, c0[(*m)*j+i], c1[(*m)*j+i], c2[(*m)*j+i], c3[(*m)*j+i], x, y, w, z);
				c0[(*m)*j+i] = p;
				c1[(*m)*j+i] = q;
				c2[(*m)*j+i] = r;
				c3[(*m)*j+i] = s;
			}
		}
	}
	return(0);
}

// qd multiplication
// double * qd
int s_m_qd(double *c0, double *c1, double *c2, double *c3, double *a, double *b0, double *b1, double *b2, double *b3, int *m, int *n, int *l)
{
	double p,q,r,s,x,y,w,z;
	int i,j,k;
	
	for (i = 0 ; i < *m ; i++) {
		for (j = 0 ; j < *n ; j++) {	
			c0[(*m)*j+i] = 0.0;
			c1[(*m)*j+i] = 0.0;
			c2[(*m)*j+i] = 0.0;
			c3[(*m)*j+i] = 0.0;
			for (k = 0 ; k < *l ; k++) {
				//multiplication
				s_mul_qd(&x, &y, &w, &z, a[(*m)*k+i], b0[(*l)*j+k], b1[(*l)*j+k], b2[(*l)*j+k], b3[(*l)*j+k]);
				//addition
				qd_add_qd(&p, &q, &r, &s, c0[(*m)*j+i], c1[(*m)*j+i], c2[(*m)*j+i], c3[(*m)*j+i], x, y, w, z);
				c0[(*m)*j+i] = p;
				c1[(*m)*j+i] = q;
				c2[(*m)*j+i] = r;
				c3[(*m)*j+i] = s;
			}
		}
	}
	return(0);
}

// qd_dd_scalar_multi_1
// qd scalar * dd matrix
int qd_dd_scalar_multi_1(double *c0, double *c1, double *c2, double *c3, double *a0, double *a1, double *a2, double *a3, double *b0, double *b1, int *m, int *n)
{
	double p10,p01,p11,p20,e00,e10,e01,e11,e20,x,y,w,z;
	int i;

	for(i=0; i<(*m)*(*n); i++) {
		c0[i] = 0.0; c1[i] = 0.0; c2[i] = 0.0; c3[i] = 0.0;
		
		//O(1) terms
		two_prod(&x, &y, a0[0], b0[i]);		c0[i] = x;	e00 = y;

		//O(eps) terms
		two_prod(&x, &y, a1[0], b0[i]);		p10 = x;	e10 = y;
		two_prod(&x, &y, a0[0], b1[i]);		p01 = x;	e01 = y;

		three_sum(&x, &y, &z, p01, p10, e00);	c1[i] = x;	p10 = y;	p01 = z;
		
		//O(eps^2) terms
		//five term sum
		two_prod(&x, &y, a1[0], b1[i]);		p11 = x;	e11 = y;
		two_prod(&x, &y, a2[0], b0[i]);		p20 = x;	e20 = y;

		three_sum(&x, &y, &z, p10, p20, p11);	p10 = x;	p20 = y;	p11 = z;

		two_sum(&x, &y, e10, e01);	e10 = x;	e01 = y;
		two_sum(&x, &y, p10, e10);	c2[i] = x;	e10 = y;
		two_sum(&x, &y, p20, e01);	p20 = x;	e01 = y;
		two_sum(&x, &y, p20, e10);	p20 = x;	e10 = y;
		p11 = p11 + e10 + e01;
		
		//O(eps^3) terms
		z = a3[0]*b0[i] + a2[0]*b1[i] + e11 + e20;
		three_sum2(&x, &y, p20, p01, z);	c3[i] = x;	e00 = y;
		p11 = p11 + e00;
  
		renormalize(&x, &y, &w, &z, c0[i], c1[i], c2[i], c3[i], p11);
		c0[i] = x;	c1[i] = y;	c2[i] = w;	c3[i] = z;
	}
	return(0);
}

// qd_dd_scalar_multi_2
// qd matrix * dd scalar
int qd_dd_scalar_multi_2(double *c0, double *c1, double *c2, double *c3, double *a0, double *a1, double *a2, double *a3, double *b0, double *b1, int *m, int *n)
{
	double p10,p01,p11,p20,e00,e10,e01,e11,e20,x,y,w,z;
	int i;

	for(i=0; i<(*m)*(*n); i++) {
		c0[i] = 0.0; c1[i] = 0.0; c2[i] = 0.0; c3[i] = 0.0;
		
		//O(1) terms
		two_prod(&x, &y, a0[i], b0[0]);		c0[i] = x;	e00 = y;

		//O(eps) terms
		two_prod(&x, &y, a1[i], b0[0]);		p10 = x;	e10 = y;
		two_prod(&x, &y, a0[i], b1[0]);		p01 = x;	e01 = y;

		three_sum(&x, &y, &z, p01, p10, e00);	c1[i] = x;	p10 = y;	p01 = z;
		
		//O(eps^2) terms
		//five term sum
		two_prod(&x, &y, a1[i], b1[0]);		p11 = x;	e11 = y;
		two_prod(&x, &y, a2[i], b0[0]);		p20 = x;	e20 = y;

		three_sum(&x, &y, &z, p10, p20, p11);	p10 = x;	p20 = y;	p11 = z;

		two_sum(&x, &y, e10, e01);	e10 = x;	e01 = y;
		two_sum(&x, &y, p10, e10);	c2[i] = x;	e10 = y;
		two_sum(&x, &y, p20, e01);	p20 = x;	e01 = y;
		two_sum(&x, &y, p20, e10);	p20 = x;	e10 = y;
		p11 = p11 + e10 + e01;
		
		//O(eps^3) terms
		z = a3[i]*b0[0] + a2[i]*b1[0] + e11 + e20;
		three_sum2(&x, &y, p20, p01, z);	c3[i] = x;	e00 = y;
		p11 = p11 + e00;
  
		renormalize(&x, &y, &w, &z, c0[i], c1[i], c2[i], c3[i], p11);
		c0[i] = x;	c1[i] = y;	c2[i] = w;	c3[i] = z;
	}
	return(0);
}

//qd multiplication
//dd * qd
int dd_m_qd(double *c0, double *c1, double *c2, double *c3, double *a0, double *a1, double *b0, double *b1, double *b2, double *b3, int *m, int *n, int *l)
{
	double p,q,r,s,x,y,w,z;
	int i,j,k;
	
	for (i = 0 ; i < *m ; i++) {
		for (j = 0 ; j < *n ; j++) {	
			c0[(*m)*j+i] = 0.0;
			c1[(*m)*j+i] = 0.0;
			c2[(*m)*j+i] = 0.0;
			c3[(*m)*j+i] = 0.0;
			for (k = 0 ; k < *l ; k++) {
				//multiplication
				dd_mul_qd(&x, &y, &w, &z, a0[(*m)*k+i], a1[(*m)*k+i], b0[(*l)*j+k], b1[(*l)*j+k], b2[(*l)*j+k], b3[(*l)*j+k]);
				//addition
				qd_add_qd(&p, &q, &r, &s, c0[(*m)*j+i], c1[(*m)*j+i], c2[(*m)*j+i], c3[(*m)*j+i], x, y, w, z);
				c0[(*m)*j+i] = p;
				c1[(*m)*j+i] = q;
				c2[(*m)*j+i] = r;
				c3[(*m)*j+i] = s;
			}
		}
	}
	return(0);
}

//qd multiplication
//qd * dd
int qd_m_dd(double *c0, double *c1, double *c2, double *c3, double *a0, double *a1, double *a2, double *a3, double *b0, double *b1, int *m, int *n, int *l)
{
	double p,q,r,s,x,y,w,z;
	int i,j,k;
	
	for (i = 0 ; i < *m ; i++) {
		for (j = 0 ; j < *n ; j++) {	
			c0[(*m)*j+i] = 0.0;
			c1[(*m)*j+i] = 0.0;
			c2[(*m)*j+i] = 0.0;
			c3[(*m)*j+i] = 0.0;
			for (k = 0 ; k < *l ; k++) {
				//multiplication
				dd_mul_qd(&x, &y, &w, &z, b0[(*l)*j+k], b1[(*l)*j+k], a0[(*m)*k+i], a1[(*m)*k+i], a2[(*m)*k+i], a3[(*m)*k+i]);
				//addition
				qd_add_qd(&p, &q, &r, &s, c0[(*m)*j+i], c1[(*m)*j+i], c2[(*m)*j+i], c3[(*m)*j+i], x, y, w, z);
				c0[(*m)*j+i] = p;
				c1[(*m)*j+i] = q;
				c2[(*m)*j+i] = r;
				c3[(*m)*j+i] = s;
			}
		}
	}
	return(0);
}

//qd multiplication
//qd * qd
int qd_m_qd(double *c0, double *c1, double *c2, double *c3, double *a0, double *a1, double *a2, double *a3, double *b0, double *b1, double *b2, double *b3, int *m, int *n, int *l)
{
	double p,q,r,s,x,y,w,z;
	int i,j,k;
	
	for (i = 0 ; i < *m ; i++) {
		for (j = 0 ; j < *n ; j++) {	
			c0[(*m)*j+i] = 0.0;
			c1[(*m)*j+i] = 0.0;
			c2[(*m)*j+i] = 0.0;
			c3[(*m)*j+i] = 0.0;
			for (k = 0 ; k < *l ; k++) {
				//multiplication
				qd_mul_qd(&x, &y, &w, &z, a0[(*m)*k+i], a1[(*m)*k+i], a2[(*m)*k+i], a3[(*m)*k+i], b0[(*l)*j+k], b1[(*l)*j+k], b2[(*l)*j+k], b3[(*l)*j+k]);
				//addition
				qd_add_qd(&p, &q, &r, &s, c0[(*m)*j+i], c1[(*m)*j+i], c2[(*m)*j+i], c3[(*m)*j+i], x, y, w, z);
				c0[(*m)*j+i] = p;
				c1[(*m)*j+i] = q;
				c2[(*m)*j+i] = r;
				c3[(*m)*j+i] = s;
			}
		}
	}
	return(0);
}


//qd multiplication
//qd(scalar) * qd(matrix)
int scalar_multi_qd(double *c0, double *c1, double *c2, double *c3, double *a0, double *a1, double *a2, double *a3, double *b0, double *b1, double *b2, double *b3, int *m, int *n)
{
	double p10,p01,p11,p20,p02,e00,e10,e01,e11,e20,e02,x,y,w,z;
	int i;

	for(i=0; i<(*m)*(*n); i++) {
		c0[i] = 0.0; c1[i] = 0.0; c2[i] = 0.0; c3[i] = 0.0;
		
		//O(1) terms
		two_prod(&x, &y, a0[0], b0[i]);	c0[i] = x;	e00 = y;

		//O(eps) terms
		two_prod(&x, &y, a0[0], b1[i]);	p01 = x;	e01 = y;
		two_prod(&x, &y, a1[0], b0[i]);	p10 = x;	e10 = y;
		three_sum(&x, &y, &z, p01, p10, e00);	
		c1[i] = x;	//O(eps) term
		p10 = y;	p01 = z;

		//O(eps^2) terms
		two_prod(&x, &y, a0[0], b2[i]);	p02 = x;	e02 = y;
		two_prod(&x, &y, a1[0], b1[i]);	p11 = x;	e11 = y;
		two_prod(&x, &y, a2[0], b0[i]);	p20 = x;	e20 = y;

		//six three sum for p10, e01, e10, p02, p11, p20
		three_sum(&x, &y, &z, p10, e01, e10);	p10 = x;	e01 = y;	e10 = z;
		three_sum(&x, &y, &z, p02, p11, p20);	p02 = x;	p11 = y;	p20 = z;
		two_sum(&x, &y, p02, p10);
		c2[i] = x;	//O(eps^2) terms
		p10 = y;
		two_sum(&x, &y, p11, e01);	p11 = x;	e01 = y;
		two_sum(&x, &y, p10, p11);	p10 = x;	p11 = y;

		//O(eps^4) terms
		e10 = e10 + p20 + e01 + p11;

		//O(eps^3) terms
		//c3[i] = a0[i]*b3[i] + a1[i]*b2[i] + a2[i]*b1[i] + a3[i]*b0[i] + e00 + e02 + e11 + e20;
		c3[i] = p10 + a0[0]*b3[i] + a1[0]*b2[i] + a2[0]*b1[i] + a3[0]*b0[i] + e02 + e11 + e20; //ë´ÇµèáóvÉ`ÉFÉbÉNÅIÅIÅI

		renormalize(&x, &y, &w, &z, c0[i], c1[i], c2[i], c3[i], e10);
		c0[i] = x;	c1[i] = y;	c2[i] = w;	c3[i] = z;
	}
	return(0);
}


//--------------------------------------------------------------------------------------------
//
// quad-double division
//
// args
// a0, a1, a2, a3, b0, b1, b2, b3 : double numbers
// a0 + a1 = dd nubmer, a0 + a1 + a2 + a3 = qd number
// *m : row
// *n : column
// return (c0,c1,c2,c3) for qd number
//
// routines
// qd_r_s  : qd - double division
// s_r_qd  : double - qd division
// qd_r_dd : qd - dd division
// dd_r_qd : dd - qd division
// qd_r_qd : qd - qd division
//
//--------------------------------------------------------------------------------------------

//qd division
//qd / double
int qd_r_s(double *c0, double *c1, double *c2, double *c3, double *a0, double *a1, double *a2, double *a3, double *b, int *m, int *n)
{
	double x,y,w,z,t0,t1,r0,r1,r2,r3,e;
	int i;
	for(i=0; i<(*m)*(*n); i++) {
		c0[i] = 0.0; c1[i] = 0.0; c2[i] = 0.0; c3[i] = 0.0;

		c0[i] = a0[i]/b[0];
		// reminder a - c_0*b
		two_prod(&x, &y, c0[i], b[0]);
		t0 = -x;
		t1 = -y;
		//qd subtruction (a - t)
		qd_add_dd(&x, &y, &w, &z, a0[i], a1[i], a2[i], a3[i], t0, t1);
		r0 = x;	r1 = y;	r2 = w;	r3 = z;

		c1[i] = r0/b[0];
		// reminder r - c_1*b
		two_prod(&x, &y, c1[i], b[0]);
		t0 = -x;
		t1 = -y;
		//qd subtruction (r - t)
		qd_add_dd(&x, &y, &w, &z, r0, r1, r2, r3, t0, t1);
		r0 = x;	r1 = y;	r2 = w;	r3 = z;

		c2[i] = r0/b[0];
		// reminder r - c_2*b
		two_prod(&x, &y, c2[i], b[0]);
		t0 = -x;
		t1 = -y;
		//qd subtruction (r - t)
		qd_add_dd(&x, &y, &w, &z, r0, r1, r2, r3, t0, t1);
		r0 = x;	r1 = y;	r2 = w;	r3 = z;

		c3[i] = r0/b[0];
		// reminder r - c_3*b
		two_prod(&x, &y, c3[i], b[0]);
		t0 = -x;
		t1 = -y;
		//qd subtruction (r - t)
		qd_add_dd(&x, &y, &w, &z, r0, r1, r2, r3, t0, t1);
		r0 = x;	r1 = y;	r2 = w;	r3 = z;
  
		e = r0/b[0];
		renormalize(&x, &y, &w, &z, c0[i], c1[i], c2[i], c3[i], e);
		c0[i] = x;
		c1[i] = y;
		c2[i] = w;
		c3[i] = z;
	}
  return(0);
}

//qd division
//double / qd
int s_r_qd(double *c0, double *c1, double *c2, double *c3, double *a, double *b0, double *b1, double *b2, double *b3, int *m, int *n)
{
	double x,y,w,z,t0,t1,t2,t3,r0,r1,r2,r3,e;
	int i;
	for(i=0; i<(*m)*(*n); i++) {
		c0[i] = 0.0; c1[i] = 0.0; c2[i] = 0.0; c3[i] = 0.0;

		c0[i] = a[i]/b0[0];
		// reminder a - c_0*b
		//multiplication
		s_mul_qd(&x, &y, &w, &z, c0[i], b0[0], b1[0], b2[0], b3[0]);
		t0 = -x;	t1 = -y;	t2 = -w;	t3 = -z;
		//qd subtruction (a - t)
		qd_add_s(&x, &y, &w, &z, t0, t1, t2, t3, a[i]);

		r0 = x;	r1 = y;	r2 = w;	r3 = z;


		c1[i] = r0/b0[0];
		// reminder r - c_1*b
		s_mul_qd(&x, &y, &w, &z, c1[i], b0[0], b1[0], b2[0], b3[0]);
		t0 = -x;	t1 = -y;	t2 = -w;	t3 = -z;
		//qd subtruction (r - t)
		qd_add_qd(&x, &y, &w, &z, r0, r1, r2, r3, t0, t1, t2, t3);
		r0 = x;	r1 = y;	r2 = w;	r3 = z;

		c2[i] = r0/b0[0];
		// reminder r - c_2*b
		s_mul_qd(&x, &y, &w, &z, c2[i], b0[0], b1[0], b2[0], b3[0]);
		t0 = -x;	t1 = -y;	t2 = -w;	t3 = -z;
		//qd subtruction (r - t)
		qd_add_qd(&x, &y, &w, &z, r0, r1, r2, r3, t0, t1, t2, t3);
		r0 = x;	r1 = y;	r2 = w;	r3 = z;

		c3[i] = r0/b0[0];
		// reminder r - c_2*b
		s_mul_qd(&x, &y, &w, &z, c3[i], b0[0], b1[0], b2[0], b3[0]);
		t0 = -x;	t1 = -y;	t2 = -w;	t3 = -z;
		//qd subtruction (r - t)
		qd_add_qd(&x, &y, &w, &z, r0, r1, r2, r3, t0, t1, t2, t3);
		r0 = x;	r1 = y;	r2 = w;	r3 = z;
  
		e = r0/b0[0];
		renormalize(&x, &y, &w, &z, c0[i], c1[i], c2[i], c3[i], e);
		c0[i] = x;
		c1[i] = y;
		c2[i] = w;
		c3[i] = z;
	}
  return(0);
}

//qd division
//qd / dd
int qd_r_dd(double *c0, double *c1, double *c2, double *c3, double *a0, double *a1, double *a2, double *a3, double *b0, double *b1, int *m, int *n)
{
	double x,y,w,z,t0,t1,r0,r1,r2,r3,e;
	int i;
	
	for(i=0; i<(*m)*(*n); i++) {
		c0[i] = 0.0; c1[i] = 0.0; c2[i] = 0.0; c3[i] = 0.0;

		c0[i] = a0[i]/b0[0];
		// reminder a - c_0*b
		//multiplication
		s_mul_dd(&x, &y, c0[i], b0[0], b1[0]);
		t0 = -x;
		t1 = -y;
		//qd subtruction (a - t)
		qd_add_dd(&x, &y, &w, &z, a0[i], a1[i], a2[i], a3[i], t0, t1);
		r0 = x;	r1 = y;	r2 = w;	r3 = z;

		c1[i] = r0/b0[0];
		// reminder r - c_1*b
		//multiplication
		s_mul_dd(&x, &y, c1[i], b0[0], b1[0]);
		t0 = -x;
		t1 = -y;
		//qd subtruction (r - t)
		qd_add_dd(&x, &y, &w, &z, r0, r1, r2, r3, t0, t1);
		r0 = x;	r1 = y;	r2 = w;	r3 = z;

		c2[i] = r0/b0[0];
		// reminder r - c_2*b
		//multiplication
		s_mul_dd(&x, &y, c2[i], b0[0], b1[0]);
		t0 = -x;
		t1 = -y;
		//qd subtruction (r - t)
		qd_add_dd(&x, &y, &w, &z, r0, r1, r2, r3, t0, t1);
		r0 = x;	r1 = y;	r2 = w;	r3 = z;

		c3[i] = r0/b0[0];
		// reminder r - c_3*b
		//multiplication
		s_mul_dd(&x, &y, c3[i], b0[0], b1[0]);
		t0 = -x;
		t1 = -y;
		//qd subtruction (r - t)
		qd_add_dd(&x, &y, &w, &z, r0, r1, r2, r3, t0, t1);
		r0 = x;	r1 = y;	r2 = w;	r3 = z;
  
		e = r0/b0[0];
		renormalize(&x, &y, &w, &z, c0[i], c1[i], c2[i], c3[i], e);
		c0[i] = x;
		c1[i] = y;
		c2[i] = w;
		c3[i] = z;
	}
  return(0);
}

//qd division
//dd / qd
int dd_r_qd(double *c0, double *c1, double *c2, double *c3, double *a0, double *a1, double *b0, double *b1, double *b2, double *b3, int *m, int *n)
{
	double x,y,w,z,t0,t1,t2,t3,r0,r1,r2,r3,e;
	int i;
	for(i=0; i<(*m)*(*n); i++) {
		c0[i] = 0.0; c1[i] = 0.0; c2[i] = 0.0; c3[i] = 0.0;

		c0[i] = a0[i]/b0[0];
		// reminder a - c_0*b
		//multiplication
		s_mul_qd(&x, &y, &w, &z, c0[i], b0[0], b1[0], b2[0], b3[0]);
		t0 = -x;	t1 = -y;	t2 = -w;	t3 = -z;
		//qd subtruction (a - t)
		qd_add_dd(&x, &y, &w, &z, t0, t1, t2, t3, a0[i], a1[i]);
		r0 = x;	r1 = y;	r2 = w;	r3 = z;

		c1[i] = r0/b0[0];
		// reminder r - c_1*b
		s_mul_qd(&x, &y, &w, &z, c1[i], b0[0], b1[0], b2[0], b3[0]);
		t0 = -x;	t1 = -y;	t2 = -w;	t3 = -z;
		//qd subtruction (r - t)
		qd_add_qd(&x, &y, &w, &z, r0, r1, r2, r3, t0, t1, t2, t3);
		r0 = x;	r1 = y;	r2 = w;	r3 = z;

		c2[i] = r0/b0[0];
		// reminder r - c_2*b
		s_mul_qd(&x, &y, &w, &z, c2[i], b0[0], b1[0], b2[0], b3[0]);
		t0 = -x;	t1 = -y;	t2 = -w;	t3 = -z;
		//qd subtruction (r - t)
		qd_add_qd(&x, &y, &w, &z, r0, r1, r2, r3, t0, t1, t2, t3);
		r0 = x;	r1 = y;	r2 = w;	r3 = z;

		c3[i] = r0/b0[0];
		// reminder r - c_2*b
		s_mul_qd(&x, &y, &w, &z, c3[i], b0[0], b1[0], b2[0], b3[0]);
		t0 = -x;	t1 = -y;	t2 = -w;	t3 = -z;
		//qd subtruction (r - t)
		qd_add_qd(&x, &y, &w, &z, r0, r1, r2, r3, t0, t1, t2, t3);
		r0 = x;	r1 = y;	r2 = w;	r3 = z;
  
		e = r0/b0[0];
		renormalize(&x, &y, &w, &z, c0[i], c1[i], c2[i], c3[i], e);
		c0[i] = x;
		c1[i] = y;
		c2[i] = w;
		c3[i] = z;
	}
  return(0);
}

//qd division
//qd / qd
int qd_r_qd(double *c0, double *c1, double *c2, double *c3, double *a0, double *a1, double *a2, double *a3, double *b0, double *b1, double *b2, double *b3, int *m, int *n)
{
	double x,y,w,z,t0,t1,t2,t3,r0,r1,r2,r3,e;
	int i;
	
	for(i=0; i<(*m)*(*n); i++) {
		c0[i] = 0.0; c1[i] = 0.0; c2[i] = 0.0; c3[i] = 0.0;

		c0[i] = a0[i]/b0[0];
		// reminder a - c_0*b
		//multiplication
		s_mul_qd(&x, &y, &w, &z, c0[i], b0[0], b1[0], b2[0], b3[0]);
		t0 = -x;	t1 = -y;	t2 = -w;	t3 = -z;
		//qd subtruction (a - t)
		qd_add_qd(&x, &y, &w, &z, a0[i], a1[i], a2[i], a3[i], t0, t1, t2, t3);
		r0 = x;	r1 = y;	r2 = w;	r3 = z;

		c1[i] = r0/b0[0];
		// reminder r - c_1*b
		//multiplication
		s_mul_qd(&x, &y, &w, &z, c1[i], b0[0], b1[0], b2[0], b3[0]);
		t0 = -x;	t1 = -y;	t2 = -w;	t3 = -z;
		//qd subtruction (r - t)
		qd_add_qd(&x, &y, &w, &z, r0, r1, r2, r3, t0, t1, t2, t3);
		r0 = x;	r1 = y;	r2 = w;	r3 = z;

		c2[i] = r0/b0[0];
		// reminder r - c_2*b
		//multiplication
		s_mul_qd(&x, &y, &w, &z, c2[i], b0[0], b1[0], b2[0], b3[0]);
		t0 = -x;	t1 = -y;	t2 = -w;	t3 = -z;
		//qd subtruction (r - t)
		qd_add_qd(&x, &y, &w, &z, r0, r1, r2, r3, t0, t1, t2, t3);
		r0 = x;	r1 = y;	r2 = w;	r3 = z;

		c3[i] = r0/b0[0];
		// reminder r - c_3*b
		//multiplication
		s_mul_qd(&x, &y, &w, &z, c3[i], b0[0], b1[0], b2[0], b3[0]);
		t0 = -x;	t1 = -y;	t2 = -w;	t3 = -z;
		//qd subtruction (r - t)
		qd_add_qd(&x, &y, &w, &z, r0, r1, r2, r3, t0, t1, t2, t3);
		r0 = x;	r1 = y;	r2 = w;	r3 = z;
  
		e = r0/b0[0];
		renormalize(&x, &y, &w, &z, c0[i], c1[i], c2[i], c3[i], e);
		c0[i] = x;
		c1[i] = y;
		c2[i] = w;
		c3[i] = z;
	}
  return(0);
}
//sloppy
/*int qd_r_qd(double *c0, double *c1, double *c2, double *c3, double *a0, double *a1, double *a2, double *a3, double *b0, double *b1, double *b2, double *b3, int *m, int *n)
{
	double x,y,w,z,t0,t1,t2,t3,r0,r1,r2,r3;//,e;
	int i;
	
	for(i=0; i<(*m)*(*n); i++) {
		c0[i] = 0.0; c1[i] = 0.0; c2[i] = 0.0; c3[i] = 0.0;
        
		c0[i] = a0[i]/b0[0];
		// reminder a - c_0*b
		//multiplication
		s_mul_qd(&x, &y, &w, &z, c0[i], b0[0], b1[0], b2[0], b3[0]);
		t0 = -x;	t1 = -y;	t2 = -w;	t3 = -z;
		//qd subtruction (a - t)
		qd_add_qd(&x, &y, &w, &z, a0[i], a1[i], a2[i], a3[i], t0, t1, t2, t3);
		r0 = x;	r1 = y;	r2 = w;	r3 = z;
        
		c1[i] = r0/b0[0];
		// reminder r - c_1*b
		//multiplication
		s_mul_qd(&x, &y, &w, &z, c1[i], b0[0], b1[0], b2[0], b3[0]);
		t0 = -x;	t1 = -y;	t2 = -w;	t3 = -z;
		//qd subtruction (r - t)
		qd_add_qd(&x, &y, &w, &z, r0, r1, r2, r3, t0, t1, t2, t3);
		r0 = x;	r1 = y;	r2 = w;	r3 = z;
        
		c2[i] = r0/b0[0];
		// reminder r - c_2*b
		//multiplication
		s_mul_qd(&x, &y, &w, &z, c2[i], b0[0], b1[0], b2[0], b3[0]);
		t0 = -x;	t1 = -y;	t2 = -w;	t3 = -z;
		//qd subtruction (r - t)
		qd_add_qd(&x, &y, &w, &z, r0, r1, r2, r3, t0, t1, t2, t3);
		r0 = x;	r1 = y;	r2 = w;	r3 = z;
        
		c3[i] = r0/b0[0];
		renormalize(&x, &y, &w, &z, c0[i], c1[i], c2[i], c3[i], 0.0);
		c0[i] = x;
		c1[i] = y;
		c2[i] = w;
		c3[i] = z;
	}
    return(0);
}*/


//--------------------------------------------------------------------------------------------
//
// renormalizing algorithm
//
// args
// a0, a1, a2, a3, a4 : double numbers
// *m : row
// *n : column
// return (b0,b1,b2,b3) for qd number(nonoverlapping)
//
//--------------------------------------------------------------------------------------------
int renorm(double *b0, double *b1, double *b2, double *b3, double *a0, double *a1, double *a2, double *a3, double *a4, int *m, int *n)
{
	double t0,t1,t2,t3,t4,s,ss,x,y;
	int i;
    
	for(i=0; i<(*m)*(*n); i++) {
		s = 0.0; ss = 0.0; t0 = 0.0; t1 = 0.0; t2 = 0.0; t3 = 0.0; t4 = 0.0;
		/*
         fast_two_sum(&x, &y, a3[i], a4[i]);
		s = x;
		t4 = y;
		fast_two_sum(&x, &y, a2[i], s);
		s = x;
		t3 = y;
		fast_two_sum(&x, &y, a1[i], s);
		s = x;
		t2 = y;
		fast_two_sum(&x, &y, a0[i], s);
		t0 = x;
		t1 = y;
		if(t1 != 0.0) {
			fast_two_sum(&x, &y, t1, t2);
			t1 = x;
			t2 = y;
			if(t2 != 0.0) {
				fast_two_sum(&x, &y,t2, t3);
				t2 = x;
				t3 = y;
				if(t3 != 0.0) {
					t3 += t4;
				} else {
					t2 += t4;
				}
			} else {
				fast_two_sum(&x, &y, t1, t3);
				t1 = x;
				t2 = y;
				if(t2 != 0.0) {
					fast_two_sum(&x, &y, t2, t4);
					t2 = x;
					t3 = y;
				} else {
					fast_two_sum(&x, &y, t1, t4);
					t1 = x;
					t2 = y;
				}
			}
		} else {
			fast_two_sum(&x, &y, t0, t2);
			t0 = x;
			t1 = y;
			if(t1 != 0.0) {
				fast_two_sum(&x, &y, t1, t3);
				t1 = x;
				t2 = y;
				if(t2 != 0.0) {
					fast_two_sum(&x, &y, t2, t4);
					t2 = x;
					t3 = y;
				} else {
					fast_two_sum(&x, &y, t1, t4);
					t1 = x;
					t2 = y;
				}
			} else {
				fast_two_sum(&x, &y, t0, t3);
				t0 = x;
				t1 = y;
				if(t1 != 0.0) {
					fast_two_sum(&x, &y, t1, t4);
					t1 = x;
					t2 = y;
				} else {
					fast_two_sum(&x, &y, t0, t4);
					t0 = x;
					t1 = y;
				}
			}
		}
		b0[i] = t0;
		b1[i] = t1;
		b2[i] = t2;
		b3[i] = t3;*/

        //[s,t4] = fast_two_sum(a4,a5);
        s = a3[i] + a4[i];
        t3 = a4[i] - (s - a3[i]);
        //[ss,t3] = fast_two_sum(a3,s);
        ss = a2[i] + s;
        t2 = s - (ss - a2[i]);
        //[s,t2] = fast_two_sum(a2,ss);
        s  = a1[i] + ss;
        t1 = ss - (s - a1[i]);
        //[b1,t1] = fast_two_sum(a1,s);
        b0[i] = a0[i] + s;
        t0 = s - (b0[i] - a0[i]);
        //[s,t3] = fast_two_sum(t3,t4);
        s = t2 + t3;
        t2 = t3 - (s - t2);
        //[ss,t2] = fast_two_sum(t2,s);
        ss = t1 + s;
        t1 = s - (ss - t1);
        //[b2,t1] = fast_two_sum(t1,ss);
        b1[i] = t0 + ss;
        t0 = ss - (b1[i] - t0);
        //[s,t2] = fast_two_sum(t2,t3);
        s = t1 + t2;
        t1 = t2 - (s -t1);
        //[b3,t1] = fast_two_sum(t1,s);
        b2[i] = t0 + s;
        t0 = s - (b2[i] - t0);
        b3[i] = t0 + t1;
	}
	return(0);
}



//--------------------------------------------------------------------------------------------
//
// internal algorithms and error free transformation
//
// routines
// fast_two_sum : s + e = a + b, s = fl(a + b), e = err(a + b), assumption |a|>|b|, Dekker('71)
// two_sum      : s + e = a + b, s = fl(a + b), e = err(a + b), Knuth('69)
// three_sum    : d + e + f = a + b + c, def are nonoverlapping, Bailey
// three_sum2   : d + e = a + b + c, de are nonoverlapping, Bailey
// two_prod     : s + e = a * b, s = fl(a * b), e = err(a * b), Verkamp and Dekker
// fast_three_accum :
// renormalize  : renormalization algorithm
// qd_add_s  : qd + double
// qd_add_dd : qd + dd
// qd_add_qd : qd + qd
// s_mul_dd  : double * dd
// s_mul_qd  : double * qd
// dd_mul_qd : dd * qd
// qd_mul_dd : qd * dd
// qd_mul_qd : qd * qd
//
//--------------------------------------------------------------------------------------------

inline void fast_two_sum(double *s, double *e, double a, double b)
{
	double v;
	s[0] = 0.0; e[0] = 0.0;

	s[0] = a + b;
	v = s[0] - a;
	e[0] = b - v;
	
	return;
}

inline void two_sum(double *s, double *e, double a, double b)
{
	double v;
	s[0] = 0.0; e[0] = 0.0;

	s[0] = a + b;
	v = s[0] - a;
	e[0] = (a - (s[0] - v)) + (b - v);
	
	return;
}

inline void three_sum(double *d, double *e, double *f, double a, double b, double c)
{
	double t1,t2,t3,v;
	d[0] = 0.0; e[0] = 0.0; f[0] = 0.0;

	t1= a + b;
	v = t1 - a;
	t2= (a - (t1 - v))+(b - v);
	
	d[0] = t1 + c;
	v = d[0] - t1;
	t3= (t1 - (d[0] - v))+(c - v);
	
	e[0] = t2 + t3;
	v = e[0] - t2;
	f[0] = (t2 - (e[0] - v))+(t3 - v);
	
	return;
}

inline void three_sum2(double *d, double *e, double a, double b, double c)
{
	double t1,t2,t3,v;
	d[0] = 0.0; e[0] = 0.0;

	t1= a + b;
	v = t1 - a;
	t2= (a - (t1 - v))+(b - v);
  
	d[0] = t1 + c;
	v = d[0] - t1;
	t3= (t1 - (d[0] - v))+(c - v);
  
	e[0] = t2 + t3;
	
	return;
}

inline void two_prod(double *p, double *e, double a, double b)
{
	double t,ah,al,bh,bl;

	p[0] = a * b;

	t = 134217729 * a;
	ah = t -(t - a);
	al = a - ah;
	t = 134217729 * b;
	bh = t -(t - b);
	bl = b - bh;

	e[0] = ((ah*bh - p[0]) + ah*bl + al*bh) + al*bl;

	return;
}

inline void fast_three_accum(double *s, double *t, double *u, double a, double b, double c)
{
    double x,y;
    bool zt,zu;
    
    two_sum(&x, &y, b, c);
    s[0] = x;    u[0] = y;
    two_sum(&x, &y, s[0], a);
    s[0] = x;    t[0] = y;

    zt = (t[0] != 0.0);
    zu = (u[0] != 0.0);
    
    if (zt && zu) { // t != 0 and u != 0 
        return;
    }
    
    if (!zu) {      // u = 0 
        u[0] = t[0];    t[0] = s[0];    s[0] = 0.0;
    } else {        // t = 0
        t[0] = s[0];    s[0] = 0.0;
    }
    return;
}

inline void renormalize(double *b0, double *b1, double *b2, double *b3, double a0, double a1, double a2, double a3, double a4)
{
	double t0,t1,t2,t3,t4,s,ss;
	s = 0.0; ss = 0.0;
	t0 = 0.0;t1 = 0.0;t2 = 0.0;t3 = 0.0;t4 = 0.0;
	/*
	fast_two_sum(&x, &y, a3, a4);
	s = x;
	t4 = y;
	fast_two_sum(&x, &y, a2, s);
	s = x;
	t3 = y;
	fast_two_sum(&x, &y, a1, s);
	s = x;
	t2 = y;
	fast_two_sum(&x, &y, a0, s);
	t0 = x;
	t1 = y;
	if(t1 != 0.0) {
		fast_two_sum(&x, &y, t1, t2);
		t1 = x;
		t2 = y;
		if(t2 != 0.0) {
			fast_two_sum(&x, &y,t2, t3);
			t2 = x;
			t3 = y;
			if(t3 != 0.0) {
				t3 += t4;
			} else {
				t2 += t4;
			}
		} else {
			fast_two_sum(&x, &y, t1, t3);
			t1 = x;
			t2 = y;
			if(t2 != 0.0) {
				fast_two_sum(&x, &y, t2, t4);
				t2 = x;
				t3 = y;
			} else {
				fast_two_sum(&x, &y, t1, t4);
				t1 = x;
				t2 = y;
			}
		}
	} else {
		fast_two_sum(&x, &y, t0, t2);
		t0 = x;
		t1 = y;
		if(t1 != 0.0) {
			fast_two_sum(&x, &y, t1, t3);
			t1 = x;
			t2 = y;
			if(t2 != 0.0) {
				fast_two_sum(&x, &y, t2, t4);
				t2 = x;
				t3 = y;
			} else {
				fast_two_sum(&x, &y, t1, t4);
				t1 = x;
				t2 = y;
			}
		} else {
			fast_two_sum(&x, &y, t0, t3);
			t0 = x;
			t1 = y;
			if(t1 != 0.0) {
				fast_two_sum(&x, &y, t1, t4);
				t1 = x;
				t2 = y;
			} else {
				fast_two_sum(&x, &y, t0, t4);
				t0 = x;
				t1 = y;
			}
		}
	}
	*/
	//[s,t4] = fast_two_sum(a4,a5);
	s = a3 + a4;
	t3 = a4 - (s - a3);
	//[ss,t3] = fast_two_sum(a3,s);
	ss = a2 + s;
	t2 = s - (ss - a2);
	//[s,t2] = fast_two_sum(a2,ss);
	s  = a1 + ss;
	t1 = ss - (s - a1);
	//[b1,t1] = fast_two_sum(a1,s);
	b0[0] = a0 + s;
	t0 = s - (b0[0] - a0);
	//[s,t3] = fast_two_sum(t3,t4);
	s = t2 + t3;
	t2 = t3 - (s - t2);
	//[ss,t2] = fast_two_sum(t2,s);
	ss = t1 + s;
	t1 = s - (ss - t1);
	//[b2,t1] = fast_two_sum(t1,ss);
	b1[0] = t0 + ss;
	t0 = ss - (b1[0] - t0);
	//[s,t2] = fast_two_sum(t2,t3);
	s = t1 + t2;
	t1 = t2 - (s -t1);
	//[b3,t1] = fast_two_sum(t1,s);
	b2[0] = t0 + s;
	t0 = s - (b2[0] - t0);
	b3[0] = t0 + t1;
	return;
}


//--------------------------------------------------------------------------------------------
//
// addition quad-double + double
//
//--------------------------------------------------------------------------------------------
inline void qd_add_s(double *c0, double *c1, double *c2, double *c3, double a0, double a1, double a2, double a3, double b)
{
	double e,x,y,w,z;
	e=0.0;
	c0[0] = 0.0; c1[0] = 0.0; c2[0] = 0.0; c3[0] = 0.0;

	two_sum(&x, &y, a0, b);	c0[0] = x;	e = y;
	two_sum(&x, &y, a1, e);	c1[0] = x;	e = y;
	two_sum(&x, &y, a2, e);	c2[0] = x;	e = y;
	two_sum(&x, &y, a3, e);	c3[0] = x;	e = y;
	renormalize(&x, &y, &w, &z, c0[0], c1[0], c2[0], c3[0], e);
	c0[0] = x;	c1[0] = y;	c2[0] = w;	c3[0] = z;
	return;
}







//--------------------------------------------------------------------------------------------
//
// addition quad-double + double-double
//
//--------------------------------------------------------------------------------------------
inline void qd_add_dd(double *c0, double *c1, double *c2, double *c3, double a0, double a1, double a2, double a3, double b0, double b1)
{
	double e1,e2,x,y,w,z;
	c0[0] = 0.0; c1[0] = 0.0; c2[0] = 0.0; c3[0] = 0.0;

	two_sum(&x, &y, a0, b0);	c0[0] = x;	e1 = y;
	two_sum(&x, &y, a1, b1);	c1[0] = x;	e2 = y;
	two_sum(&x, &y, c1[0], e1);	c1[0] = x;	e1 = y;
  	two_sum(&x, &y, a2, e2);	c2[0] = x;	e2 = y;
	two_sum(&x, &y, c2[0], e1);	c2[0] = x;	e1 = y;
	two_sum(&x, &y, e1, e2);	e1 = x;		e2 = y;
	two_sum(&x, &y, a3, e1);	c3[0] = x;	e1 = y;
	e1 = e1 + e2;
	renormalize(&x, &y, &w, &z, c0[0], c1[0], c2[0], c3[0], e1);
	c0[0] = x;	c1[0] = y;	c2[0] = w;	c3[0] = z;
	return;
}

//--------------------------------------------------------------------------------------------
//
// addition quad-double + quad-double
//
//--------------------------------------------------------------------------------------------
inline void qd_add_qd(double *c0, double *c1, double *c2, double *c3, double a0, double a1, double a2, double a3, double b0, double b1, double b2, double b3)
{
	double e1,e2,e3,e4,x,y,w,z;
	c0[0] = 0.0; c1[0] = 0.0; c2[0] = 0.0; c3[0] = 0.0;

	two_sum(&x, &y, a0, b0);	c0[0] = x;	e1 = y;
	two_sum(&x, &y, a1, b1);	c1[0] = x;	e2 = y;
	two_sum(&x, &y, c1[0], e1);	c1[0] = x;	e1 = y;
	two_sum(&x, &y, a2, b2);	c2[0] = x;	e3 = y;
  	three_sum(&x, &y, &z, c2[0], e2, e1);	c2[0] = x;	e1 = y;	e2 = z;
	two_sum(&x, &y, a3, b3);	c3[0] = x;	e4 = y;
	three_sum2(&x, &y, c3[0], e3, e1);		c3[0] = x;	e1 = y;
	e1 = e1 + e2 + e4;
	renormalize(&x, &y, &w, &z, c0[0], c1[0], c2[0], c3[0], e1);
	c0[0] = x;	c1[0] = y;	c2[0] = w;	c3[0] = z;
	return;
}

// double - DD multiplication
inline void s_mul_dd(double *chi, double *clo, double a, double bhi, double blo)
{
	double p1,p2,t,ah,al,bh,bl;

	chi[0] = 0.0;	clo[0] = 0.0;

	t = 134217729 * a;	ah = t - (t - a);	al = a - ah;
	t = 134217729 * bhi;bh = t - (t - bhi);	bl = bhi - bh;
	
	p1 = a * bhi;
	p2 = ((ah * bh - p1) + ah * bl + al * bh) + al * bl;
	p2 = p2 + a * blo;
	
	chi[0] = p1 + p2;
	clo[0] = p2 - (chi[0] - p1);
}

// double - QD multiplication
inline void s_mul_qd(double *c0, double *c1, double *c2, double *c3, double b, double a0, double a1, double a2, double a3)
{
	double e0,e1,e2,x,y,w,z;
	c0[0] = 0.0; c1[0] = 0.0; c2[0] = 0.0; c3[0] = 0.0;
		
	two_prod(&x, &y, a0, b);	c0[0] = x;	e0 = y;
	two_prod(&x, &y, a1, b);	c1[0] = x;	e1 = y;
	two_sum(&x, &y, c1[0], e0);	c1[0] = x;	e0 = y;
	two_prod(&x, &y, a2, b);	c2[0] = x;	e2 = y;
	three_sum(&x, &y, &z, c2[0], e1, e0);	c2[0] = x;	e0 = y;	e1 = z;
	c3[0] = a3*b;
	three_sum2(&x, &y, c3[0], e2, e0);		c3[0] = x;	e0 = y;
	e0 = e0 + e1;
	
	renormalize(&x, &y, &w, &z, c0[0], c1[0], c2[0], c3[0], e0);
	c0[0] = x;	c1[0] = y;	c2[0] = w;	c3[0] = z;
	return;
}


// DD - QD multiplication
inline void qd_mul_dd(double *c0, double *c1, double *c2, double *c3, double a0, double a1, double a2, double a3, double b0, double b1)
{
	double p10,p01,p11,p20,e00,e10,e01,e11,e20,x,y,w,z;
	c0[0] = 0.0; c1[0] = 0.0; c2[0] = 0.0; c3[0] = 0.0;
		
	//O(1) terms
	two_prod(&x, &y, a0, b0);		c0[0] = x;	e00 = y;

	//O(eps) terms
	two_prod(&x, &y, a1, b0);		p10 = x;	e10 = y;
	two_prod(&x, &y, a0, b1);		p01 = x;	e01 = y;

	three_sum(&x, &y, &z, p01, p10, e00);	c1[0] = x;	p10 = y;	p01 = z;
		
	//O(eps^2) terms
	//five term sum
	two_prod(&x, &y, a1, b1);		p11 = x;	e11 = y;
	two_prod(&x, &y, a2, b0);		p20 = x;	e20 = y;

	three_sum(&x, &y, &z, p10, p20, p11);	p10 = x;	p20 = y;	p11 = z;

	two_sum(&x, &y, e10, e01);	e10 = x;	e01 = y;
	two_sum(&x, &y, p10, e10);	c2[0] = x;	e10 = y;
	two_sum(&x, &y, p20, e01);	p20 = x;	e01 = y;
	two_sum(&x, &y, p20, e10);	p20 = x;	e10 = y;
	p11 = p11 + e10 + e01;
		
	//O(eps^3) terms
	z = a3*b0 + a2*b1 + e11 + e20;
	three_sum2(&x, &y, p20, p01, z);	c3[0] = x;	e00 = y;
	p11 = p11 + e00;
  
	renormalize(&x, &y, &w, &z, c0[0], c1[0], c2[0], c3[0], p11);
	c0[0] = x;	c1[0] = y;	c2[0] = w;	c3[0] = z;
	return;
}

// DD - QD multiplication
inline void dd_mul_qd(double *c0, double *c1, double *c2, double *c3, double b0, double b1, double a0, double a1, double a2, double a3)
{
	double p10,p01,p11,p20,e00,e10,e01,e11,e20,x,y,w,z;
	c0[0] = 0.0; c1[0] = 0.0; c2[0] = 0.0; c3[0] = 0.0;
		
	//O(1) terms
	two_prod(&x, &y, a0, b0);	c0[0] = x;	e00 = y;

	//O(eps) terms
	two_prod(&x, &y, a1, b0);	p10 = x;	e10 = y;
	two_prod(&x, &y, a0, b1);	p01 = x;	e01 = y;
	three_sum(&x, &y, &z, p01, p10, e00);	c1[0] = x;	p10 = y;	p01 = z;
		
	//O(eps^2) terms
	//five term sum
	two_prod(&x, &y, a1, b1);	p11 = x;	e11 = y;
	two_prod(&x, &y, a2, b0);	p20 = x;	e20 = y;
	three_sum(&x, &y, &z, p10, p20, p11);	p10 = x;	p20 = y;	p11 = z;
	two_sum(&x, &y, e10, e01);	e10 = x;	e01 = y;
	two_sum(&x, &y, p10, e10);	c2[0] = x;	e10 = y;
	two_sum(&x, &y, p20, e01);	p20 = x;	e01 = y;
	two_sum(&x, &y, p20, e10);	p20 = x;	e10 = y;
	p11 = p11 + e10 + e01;
		
	//O(eps^3) terms
	z = a3*b0 + a2*b1 + e11 + e20;
	three_sum2(&x, &y, p20, p01, z);	c3[0] = x;	e00 = y;
	p11 = p11 + e00;
  
	renormalize(&x, &y, &w, &z, c0[0], c1[0], c2[0], c3[0], p11);
	c0[0] = x;	c1[0] = y;	c2[0] = w;	c3[0] = z;
	return;
}


// scalar qd multi
inline void qd_mul_qd(double *c0, double *c1, double *c2, double *c3, double a0, double a1, double a2, double a3, double b0, double b1, double b2, double b3)
{
	double p10,p01,p11,p20,p02,e00,e10,e01,e11,e20,e02,x,y,w,z;
	c0[0] = 0.0; c1[0] = 0.0; c2[0] = 0.0; c3[0] = 0.0;
		
	//O(1) terms
	two_prod(&x, &y, a0, b0);
	c0[0] = x;	//O(1) terms
	e00 = y;

	//O(eps) terms
	two_prod(&x, &y, a0, b1);
	p01 = x;
	e01 = y;

	two_prod(&x, &y, a1, b0);
	p10 = x;
	e10 = y;

	three_sum(&x, &y, &z, p01, p10, e00);
	c1[0] = x;	//O(eps)
	p10 = y;	//O(eps^2)
	p01 = z;	//O(eps^3)

	//O(eps^2) terms
	two_prod(&x, &y, a0, b2);
	p02 = x;
	e02 = y;

	two_prod(&x, &y, a1, b1);
	p11 = x;
	e11 = y;

	two_prod(&x, &y, a2, b0);
	p20 = x;
	e20 = y;

	//six three sum for p10, e01, e10, p02, p11, p20
	three_sum(&x, &y, &z, p10, e01, e10);
	p10 = x;
	e01 = y;
	e10 = z;

	three_sum(&x, &y, &z, p02, p11, p20);
	p02 = x;
	p11 = y;
	p20 = z;

	two_sum(&x, &y, p02, p10);
	c2[0] = x;	//O(eps^2) terms
	p10 = y;

	two_sum(&x, &y, p11, e01);
	p11 = x;
	e01 = y;

	two_sum(&x, &y, p10, p11);
	p10 = x;	//O(eps^3) terms
	p11 = y;

	e10 = e10 + p20 + e01 + p11;	//O(eps^4) terms
		
	//O(eps^3) terms
	c3[0] = p01 + p10 + a0*b3 + a1*b2 + a2*b1 + a3*b0 + e02 + e11 + e20;

	renormalize(&x, &y, &w, &z, c0[0], c1[0], c2[0], c3[0], e10);
	c0[0] = x;
	c1[0] = y;
	c2[0] = w;
	c3[0] = z;
	return;
}